#!/usr/bin/env python3
##############################################################################
#
#    MyGNUHealth : Mobile and Desktop PHR node for GNU Health
#
#           MyGNUHealth is part of the GNU Health project
#
##############################################################################
#
#    GNU Health: The Libre Digital Health Ecosystem
#    Copyright (C) 2008-2024 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2024 GNU Solidario <health@gnusolidario.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import sys
import os
import logging
import kivy
import mygnuhealth.about as about
from mygnuhealth.user_account import UserAccount

from kivy.app import App
from kivy.clock import Clock
from kivy.uix.label import Label
from kivy.uix.button import Button
from kivy.uix.screenmanager import Screen, ScreenManager
from kivy.core.window import Window
from kivy.lang import Builder
from kivy.utils import get_color_from_hex
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.scrollview import ScrollView
from kivy.uix.popup import Popup
from kivy.uix.image import Image
from kivy.properties import (ObjectProperty, ListProperty,
                             DictProperty, StringProperty)

from mygnuhealth.core import CodeNameMap
from mygnuhealth.lang import tr
# Import some of the screens from their own module files
from mygnuhealth.tracker_bio_cardio import TrackerBioCardioScreen as cardio
from mygnuhealth.tracker_bio_glucose import TrackerBioGlucoseScreen as glucose
from mygnuhealth.tracker_bio_weight import TrackerBioWeightScreen as weight
from mygnuhealth.tracker_bio_osat import TrackerBioOsatScreen as osat

from mygnuhealth.tracker_lifestyle_pactivity import \
    TrackerLifestylePactivityScreen as pactivity
from mygnuhealth.tracker_lifestyle_nutrition import \
    TrackerLifestyleNutritionScreen as nutrition

from mygnuhealth.tracker_lifestyle_sleep import TrackerLifestyleSleepScreen \
    as sleep

from mygnuhealth.tracker_lifestyle_social import TrackerLifestyleSocialScreen \
    as social_activity

from mygnuhealth.tracker_psycho_mood import TrackerPsychoMoodScreen \
    as mood

from mygnuhealth.profile_settings import ProfileSettings as profile

from mygnuhealth.network_settings import NetworkSettings as network

from mygnuhealth.bluetooth_settings import BluetoothSettings as bluetooth

from mygnuhealth.book_of_life import BookofLife as bol

from mygnuhealth.page_of_life import PoL as pol

kivy.require('2.3.0')


""" By default Kivy looks for the file name that
    of the same as the main class (MyGNUHealthApp) in
    lower case and without the App suffix, "so mygnuhealth.kv"

    We want to place the user interface in the ui directory, so
    we'll use the Builder load_file method for that.
"""
moddir = os.path.dirname(os.path.abspath(__file__))

# Change directory to this module cwd so we can invoke the images
# and other files relative to it.
os.chdir(moddir)


class Menu(BoxLayout):
    manager = ObjectProperty(None)


# Declare the screens
class InitialScreen(Screen):
    """ In this initial class, MyGNUHealth checks if the
        user account has been created. If that is the case
        the next screen will be the login screen.
        If the user does not exist, it will take them to the
        user account initial setup wizard.
    """

    def account_status(self):
        acct = UserAccount()
        account = acct.account_exist()
        if account:
            self.manager.current = "login"
        else:
            self.manager.current = "newuser"


class LoginScreen(Screen):

    personal_key = ObjectProperty()

    def validate_credentials(self):
        # Check userid
        acct = UserAccount()
        if acct.login(self.personal_key.text):
            logging.info("Welcome to the jungle!")
            App.get_running_app().login_status = True
            self.manager.current = "phr"

        else:
            # switching the current screen to display validation result
            logging.error("Wrong key")

            popup = Popup(
                title=tr._('Invalid Credentials'),
                content=Label(
                    text=tr._('Wrong Personal key')),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

            # reset the personal key
            self.personal_key.text = ""


class NewUserScreen(Screen):

    sex_map = CodeNameMap(
        mapinfo=[
            {'code': 'Female', 'name': tr.N_('Female')},
            {'code': 'Male', 'name': tr.N_('Male')},
            {'code': 'Other', 'name': tr.N_('Other')},
        ])

    def init_user(self, username, sex_name, height, bday,
                  bmonth, byear, pkey, pkey_repeat):
        birthdate = [byear, bmonth, bday]
        acct = UserAccount()

        sex_code = NewUserScreen.sex_map.get_code(sex_name)

        if (acct.createAccount(pkey, pkey_repeat,
                               height, username, birthdate, sex_code)):
            self.manager.current = "login"


class AboutScreen(Screen):
    myghinfo = about


class PHRScreen(Screen):
    pass


class HealthTrackerScreen(Screen):
    pass


class ProfileSettingsScreen(Screen):
    person_height = ObjectProperty()
    person_language = ObjectProperty()
    person_fedacct = ObjectProperty()

    person_profile = profile

    def on_pre_enter(self):
        # Get the default values for height and federation account
        self.person_height = profile.default_height()
        self.person_language = profile.default_language_name()
        self.person_fedacct = profile.default_fedacct()


class NetworkSettingsScreen(Screen):
    # Default information of the thalamus server
    thalamus = DictProperty({
        'federation_account': '',
        'protocol': 'https',
        'federation_server': '',
        'federation_port': 8443,
        'enable_sync': False})

    network_settings = network

    def on_pre_enter(self):
        # Get the values from the current network settings
        if (self.network_settings.fedinfo):
            self.thalamus = self.network_settings.fedinfo


class BluetoothSettingsScreen(Screen):

    bluetooth_settings = bluetooth

    def read_device_info(self, model):
        """Retrieves the device features from the local database"""
        dev_info = self.bluetooth_settings.get_device_info(model)

        if dev_info:
            self.ids['address'].text = dev_info['address']

            for feature in dev_info['features'].keys():
                self.ids[feature].state = dev_info['features'][feature]

    def discover_device(self, model):
        """ Assign to the TextInput with id address the UID address
            of the device referenced by the model
        """
        self.ids.address.text = \
            self.bluetooth_settings.get_device_addr(model) or ''
        logging.info(f"Device address for {model}  {self.ids.address.text}")

    def sync_device(self, addr, model, features):
        """ Retrieves and saves to the local database the values of the
            services associated to the device address
        """
        active_features = []
        for feature, value in features.items():
            # In Kivy, the ToggleButton state is 'down' when pressed (selected)
            if value == 'down':
                active_features.append(feature)

        for feature in active_features:
            measure = self.bluetooth_settings.get_measures(
                addr, model, feature)
            if measure and feature == 'hr':
                logging.info(f"Retrieved measure: {measure}")
                logging.info("Syncing the results...")
                sync_info = f'synced from {model}'
                cardio.set_values(
                    self, systolic=-1, diastolic=-1,
                    heart_rate=measure, extra_info=sync_info)


class TrackerBioScreen(Screen):
    bp = ListProperty(["", "", "", ""])  # Date, sys, dia and hr
    glucose = ListProperty(["", ""])     # Date and glycemia value
    weight = ListProperty(["", ""])     # Date and weight value
    osat = ListProperty(["", ""])     # Date and osat value

    def on_pre_enter(self):
        # Refresh the chart anytime we access the bio summary
        self.bp = cardio.getBP()
        self.glucose = glucose.getGlucose()
        self.weight = weight.getWeight()
        self.osat = osat.getOsat()


class PageofLifeScreen(Screen):
    pdate = ListProperty(["", "", "", "", ""])  # Year, month, day, hour, min
    domain_names = ListProperty()
    domain_context_names = ListProperty()
    relevance_names = ListProperty()

    rsinfo = DictProperty(
        {'dbsnp': '', 'gene': '', 'protein': '', 'variant': '',
         'aa_change': '', 'category': '', 'disease': ''})

    page = pol

    def on_pre_enter(self):
        self.pdate = pol.get_date()
        self.domain_names = pol.get_domain_names()
        self.relevance_names = pol.get_relevance_names()

    def get_domain_context_names(self, domain_name):
        self.domain_context_names = pol.get_context_names(domain_name)

    def domain_context_is_genetic(self, domain_name):
        return pol.domain_context_is_genetic(domain_name)

    def get_rsinfo(self, rs):
        if (rs):
            res = pol.checkSNP(rs)
            if res:
                self.rsinfo = res
            else:
                # Reset to '' all the rsinfo dictionary key values
                self.rsinfo = dict.fromkeys(self.rsinfo, '')


class ScrollableLabel(ScrollView):
    text = StringProperty('')


class BookofLifeSummaryButton(Button):
    text = StringProperty('')
    page_info = DictProperty({})


class BookofLifeScreen(Screen):
    bolgrid = ObjectProperty()
    book = ListProperty([])

    bookcls = bol

    pols_cache = []
    last_group = 1
    group_range = (0, 0)
    group_length = 50

    clock = None
    syncing_image = None

    def summary_button_callback(self, event):
        page_info = event.page_info
        popup = Popup(
            title=page_info['domain'],
            content=ScrollableLabel(
                text=page_info['summary']),
            size_hint=(0.7, 0.7),
            auto_dismiss=True)
        popup.open()

    def create_pols_widgets(self, group=1, use_pols_cache=False):
        if use_pols_cache:
            pols = self.pols_cache
        else:
            pols = list(reversed(bol.read_book()))
            self.pols_cache = pols

        group_range = (1, len(pols) // self.group_length + 1)
        group = min(group_range[1], max(group, 1))

        self.group_range = group_range
        self.last_group = group

        begin = (group - 1) * self.group_length
        to = group * self.group_length

        pols_showed = pols[begin:to]

        widgets = []

        for page in pols_showed:
            if page["fsynced"]:
                tags = ''
            elif page['privacy']:
                tags = tr._('Private')
            else:
                tags = tr._('Not synced')

            if tags != '':
                tags = '\n(' + tags + ')'

            page_date = (f'[color=#32393a]{str(page["date"])}' +
                         f'[b]{tags}[/b][/color]')
            page_domain_summary = (
                f'[color=#32393a][b]{str(page["domain"])}:[/b]\n' +
                f'{str(page["summary"])}[/color]')

            date_widget = Label(text=page_date, markup=True)

            domain_summary_widget = BookofLifeSummaryButton(
                text=page_domain_summary,
                page_info=page)
            domain_summary_widget.bind(
                on_press=self.summary_button_callback)

            widgets.append(
                (date_widget, domain_summary_widget))

        return widgets

    def on_pre_enter(self):
        widgets = self.create_pols_widgets(group=1)
        self.bolgrid_add_widgets(widgets)

    def bolgrid_add_widgets(self, widgets):
        self.bolgrid.clear_widgets()

        for date_widget, domain_summary_widget in widgets:
            self.bolgrid.add_widget(date_widget)
            self.bolgrid.add_widget(domain_summary_widget)

        self.ids.show_pols_groups_info.text = self.show_pols_groups_info()

    def show_pols_groups_info(self):
        group_range = self.group_range
        last_group = self.last_group

        if last_group == group_range[0]:
            fmt = '{0}[{1}] .. {3}{4}'
        elif last_group == group_range[1]:
            fmt = '{0}{1} .. [{3}]{4}'
        else:
            fmt = '{0}{1} .. [{2}] .. {3}{4}'

        info = fmt.format(
            '[color=#32393a]',
            group_range[0],
            last_group,
            group_range[1],
            '[/color]')

        return info

    def show_next_pols_group(self, arg=1):
        group = self.last_group + arg
        self.jump_to_pols_group(group)

    def show_previous_pols_group(self, arg=1):
        self.show_next_pols_group(-1 * arg)

    def jump_to_pols_group(self, group=1):
        try:
            group = int(group)
        except BaseException:
            group = 1
        widgets = self.create_pols_widgets(
            group=group, use_pols_cache=True)
        self.bolgrid_add_widgets(widgets)
        self.ids.bolscroll.scroll_y = 1

    def sync_book(self, fedkey, button_id):
        sync_button = self.ids[button_id]
        syncing_image = Image(
            source='./images/syncing-icon.gif',
            anim_delay=0.1,
            size=sync_button.size,
            pos=sync_button.pos)
        sync_button.add_widget(syncing_image)
        self.syncing_image = syncing_image

        def redraw(self, args):
            syncing_image.size = sync_button.size
            syncing_image.pos = sync_button.pos
        self.bind(size=redraw, pos=redraw)

        callback = self.refresh_current_pols_group
        self.bookcls.sync_book(fedkey, callback=callback)

    def refresh_current_pols_group(self, sync_state):
        pols = list(reversed(bol.read_book()))
        self.pols_cache = pols
        if self.clock:
            self.clock.cancel()
        self.clock = Clock.schedule_once(
            lambda dt: self.refresh_current_pols_group_for_clock(sync_state),
            0.5)

    def refresh_current_pols_group_for_clock(self, sync_state):
        sync_error = sync_state.get('error')
        if sync_error:
            popup = Popup(
                title=tr._('Federation Sync Error'),
                content=ScrollableLabel(
                    text=str(sync_error)),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()
        else:
            self.jump_to_pols_group(self.last_group),

        sync_button = self.syncing_image.parent
        sync_button.remove_widget(self.syncing_image)


class TrackerLifestyleScreen(Screen):
    # Date, aerobic, anaerobic, steps
    pactivity = ListProperty(["", "", "", ""])
    # Date, morning, afternoon, evening, total, info
    nutrition = ListProperty(["", "", "", "", ""])
    # Date, sleep_time, quality, info
    sleep = ListProperty(["", "", "", ""])
    # Date, meaningful social activities time, info
    social_activity = ListProperty(["", "", ""])

    def on_pre_enter(self):
        # Refresh the chart anytime we access the lifestyle summary
        self.pactivity = pactivity.getPA()
        self.nutrition = nutrition.getNutrition()
        self.sleep = sleep.getSleep()
        self.social_activity = social_activity.getSA()


class TrackerPsychoScreen(Screen):
    # Date, mood, energy, info
    mood = ListProperty(["", "", "", ""])

    def on_pre_enter(self):
        # Refresh the chart anytime we access the lifestyle summary
        self.mood = mood.getMood()


class ScreenController(ScreenManager):
    pass


# Load the main kv file from the UI directory
# call load_file from here, after all the classes are declared
kv = Builder.load_file('ui/mygnuhealth.kv')

""" By default Kivy looks for the file name that
    of the same as the main class (MyGNUHealthApp) in
    lower case and without the "App" suffix, "so mygnuhealth.kv"
    We use the Builder load_file method to call it from the ui dir
"""


class MyGnuHealthApp(App):
    # Use a soft background color
    # Window.clearcolor = get_color_from_hex('#f5fafa')
    Window.clearcolor = get_color_from_hex('#ffffff')

    """ The last_known_screen keeps the latest loaded screen.
        We use this variable for return to the previous screen
        as in the the About page (similar to a stack pop operation)
    """
    last_known_screen = None
    login_status = False

    language_code = profile.default_language_code()
    tr.switch_language(language_code)

    def build(self):
        self.title = f"MyGNUHealth {about.__version__}"
        self.icon = 'images/mygnuhealth-icon.png'
        return Menu()

    def bailout(self, rc):
        """ Exit the application with the given return code, rc
        """
        sys.exit(rc)


def mygh():
    MyGnuHealthApp().run()


if __name__ == "__main__":
    mygh()



##############################################################################
#
#    MyGNUHealth : Mobile and Desktop PHR node for GNU Health
#
#           MyGNUHealth is part of the GNU Health project
#
##############################################################################
#
#    GNU Health: The Libre Digital Health Ecosystem
#    Copyright (C) 2008-2024 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2024 GNU Solidario <health@gnusolidario.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from kivy.uix.label import Label
from kivy.uix.popup import Popup
from mygnuhealth.core import maindb
from mygnuhealth.lang import tr
from mygnuhealth.fedlogin import test_federation_connection as fc


class NetworkSettings():
    fed_table = maindb.table('federation')
    fedinfo = {}

    # Cast to dict the resulting tinydb.table.Document
    if (len(fed_table) > 0):
        fedinfo = dict(fed_table.all()[0])

    def update_federation_info(protocol, federation_server,
                               federation_port, enable_sync):

        # If the "Singleton" table is empty, insert, otherwise, update
        # TODO: Use upsert with doc_id == 1 as condition
        if not len(NetworkSettings.fed_table):
            NetworkSettings.fed_table.insert(
                {'protocol': protocol,
                 'federation_server': federation_server,
                 'federation_port': federation_port,
                 'enable_sync': enable_sync})
        else:
            NetworkSettings.fed_table.update(
                {'protocol': protocol,
                 'federation_server': federation_server,
                 'federation_port': federation_port,
                 'enable_sync': enable_sync})

    def test_connection(protocol, *args):
        conn_res = fc(protocol, *args)

        if (conn_res == 0):
            msg = tr._("Connection OK")
        if (conn_res == -1):
            msg = tr._("Invalid Credentials")
        if (conn_res == -2):
            msg = tr._("Network Error")
        if (conn_res == -3):
            msg = tr._("Unknown error")

        status_msg = f"{msg} (rc:{conn_res})"
        popup = Popup(
            title=tr._('Federation Connection Status'),
            content=Label(text=status_msg),
            size_hint=(0.5, 0.5), auto_dismiss=True)
        popup.open()

        return [conn_res, msg]



#!/usr/bin/env python3
####################################################################
#   Copyright (C) 2020-2024 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2024 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

from mygnuhealth.__init__ import __version__ as myghver
import os

__version__ = myghver

__appname__ = "MyGNUHealth"

__description__ = "The GNU Health Personal Health Record"

__homepage__ = "https://www.gnuhealth.org"

__author__ = "Luis Falcon"

__email__ = "info@gnuhealth.org"

__download_url__ = 'https://ftp.gnu.org/gnu/health/mygnuhealth'

__license__ = "GPL v3+"
__copyright__ = "Copyright 2008-2024 GNU Solidario"

__thanks__ = open(
    os.path.join(os.path.dirname(__file__), "contributors.txt")).read()



##############################################################################
#
#    MyGNUHealth : Mobile and Desktop PHR node for GNU Health
#
#           MyGNUHealth is part of the GNU Health project
#
##############################################################################
#
#    GNU Health: The Libre Digital Health Ecosystem
#    Copyright (C) 2008-2024 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2024 GNU Solidario <health@gnusolidario.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import bcrypt
import locale
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from mygnuhealth.lang import tr
from mygnuhealth.core import (get_personal_key, get_user_profile,
                              get_federation_account,
                              maindb, CodeNameMap)


class ProfileSettings():

    def check_current_password(current_password):
        personal_key = get_personal_key(maindb)
        cpw = current_password.encode()
        rc = bcrypt.checkpw(cpw, personal_key)
        if not rc:
            popup = Popup(
                title=tr._('Wrong password'),
                content=Label(
                    text=tr._("Current password does not match")),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

        return rc

    def check_new_password(password, password_repeat):
        rc = None
        print(len(password))
        if ((password == password_repeat) and (len(password) > 2)):
            rc = password
        if not rc:
            popup = Popup(
                title=tr._('Wrong values'),
                content=Label(
                    text=tr._("Passwords don't match or key is too small")),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()
        return rc

    def update_personalkey(password):
        encrypted_key = bcrypt.hashpw(password.encode('utf-8'),
                                      bcrypt.gensalt()).decode('utf-8')

        credentialstable = maindb.table('credentials')
        if (len(credentialstable) > 0):
            credentialstable.update({'personal_key': encrypted_key})
        else:
            print("Initializing credentials table")
            credentialstable.insert({'personal_key': encrypted_key})

        print("Saved personal key", encrypted_key)

    def update_height(profile):
        profiletable = maindb.table('profile')
        if (len(profiletable) > 0):
            print(f"Updating height to {profile['height']}")
            profiletable.update({'height': profile['height']})

        else:
            print(f"Initializing profile. Setting height {profile['height']}")
            profiletable.insert({'height': profile['height']})

        popup = Popup(
            title=tr._('Success!'),
            content=Label(
                text=tr._("Height succesfully updated")),
            size_hint=(0.5, 0.5), auto_dismiss=True)
        popup.open()

        return True

    def set_height(height):
        profile_height = {'height': height}
        if (height):
            ProfileSettings.update_height(profile_height)

    def all_languages_info():
        languages = [
            {'code': 'en', 'name': 'English (en)'},
            {'code': 'fr', 'name': 'Français (fr)'},
            {'code': 'de', 'name': 'Deutsch (de)'},
            {'code': 'es', 'name': 'Español (es)'},
            {'code': 'zh_CN', 'name': '中文简体 (zh_CN)'},
        ]
        return languages

    languages_map = CodeNameMap(
        mapfunc=all_languages_info)

    def all_languages():
        languages_map = ProfileSettings.languages_map
        return languages_map.get_names()

    def set_language(language_name):
        languages_map = ProfileSettings.languages_map
        languages_map.update_history(language_name)
        language_code = languages_map.get_code(language_name)
        profile_language = {'language': language_code}
        if (language_code):
            ProfileSettings.update_language(profile_language)
            tr.switch_language(language_code)

    def update_language(profile):
        profiletable = maindb.table('profile')
        if (len(profiletable) > 0):
            print(f"Updating language to {profile['language']}")
            profiletable.update({'language': profile['language']})

        else:
            print("Initializing profile. " +
                  f"Setting language {profile['language']}")
            profiletable.insert({'language': profile['language']})

        popup = Popup(
            title=tr._('Success!'),
            content=Label(
                text=tr._("Language succesfully updated")),
            size_hint=(0.5, 0.5), auto_dismiss=True)
        popup.open()

        return True

    def update_fedacct(fedacct):
        fedtable = maindb.table('federation')
        if (len(fedtable) > 0):
            fedtable.update({'federation_account': fedacct})
        else:
            print("Initializing federation settings")
            fedtable.insert({'federation_account': fedacct})

        popup = Popup(
            title=tr._('Success!'),
            content=Label(
                text=tr._("Federation account succesfully updated")),
            size_hint=(0.5, 0.5), auto_dismiss=True)
        popup.open()

        return True

    def set_fedacct(userfedacct):
        if (userfedacct):
            ProfileSettings.update_fedacct(userfedacct)

    def validate_pkey_update(current_password, password,
                             password_repeat):
        if (ProfileSettings.check_current_password(current_password) and
                ProfileSettings.check_new_password(password, password_repeat)):
            print("Pkey validation OK... updating personal key")
            ProfileSettings.update_personalkey(password)
            popup = Popup(
                title=tr._('Success!'),
                content=Label(
                    text=tr._("Personal Key sucessfully updated")),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

        else:
            print("Pkey validation error")

    def default_height():
        if get_user_profile(maindb):
            return get_user_profile(maindb)['height']

    def default_language_name():
        languages_map = ProfileSettings.languages_map
        language_code = ProfileSettings.default_language_code()
        return (languages_map.get_name(language_code) or
                languages_map.get_name('en'))

    def default_language_code():
        if get_user_profile(maindb):
            language_code = (get_user_profile(maindb).get('language')
                             or locale.getdefaultlocale()[0]
                             or 'en')
            return language_code
        else:
            return locale.getdefaultlocale()[0] or 'en'

    def default_fedacct():
        return get_federation_account()



#  mygh_import_readings.py
#
#  Copyright 2023 Luis Falcon <falcon@gnuhealth.org>
#  Copyright 2023 GNU Solidario <health@gnusolidario.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import csv
import sys
import datetime
from uuid import uuid4
from tinydb import TinyDB
from mygnuhealth.myghconf import dbfile
db = TinyDB(dbfile)


def save_values(table, vals):
    """
        Insert the record on the appropriate table
    """
    dbtable = db.table(table)
    print("Inserting into DB", dbtable, vals)
    dbtable.insert(vals)

    print("Saved event", vals)


def set_weight(vals):
    """
        Validate weight (kg) and set the value
    """
    weight = vals['weight']
    height = 0
    if vals['height']:
        height = int(vals['height'])

    if 20 <= weight <= 300:
        # If the height is included, calculate the BMI
        if height in range(60, 250):
            bmi = weight / ((height / 100)**2)
            bmi = round(bmi, 1)  # Use one decimal
            vals['bmi'] = bmi

        save_values('weight', vals)


def set_bp(vals):
    """
        Validate sanity of BP readings
    """
    systolic = vals['systolic']
    diastolic = vals['diastolic']

    if systolic in range(40, 300) and diastolic in range(30, 200):
        save_values('bloodpressure', vals)


def set_hr(vals):
    """
        Validate sanity of HR readings
    """
    hr = vals['heart_rate']
    if hr in range(20, 300):
        save_values('heart_rate', vals)


def validate_date(vdate, vtime):
    """
    Validate the given date and time and return ISO formated value
    """
    date_string = f"{vdate} {vtime}"
    try:
        stodate = datetime.datetime.strptime(date_string,
                                             '%Y-%m-%d %H:%M').isoformat()
    except BaseException:
        print("Error on the date or time. Skipping...")

    return (stodate)


def parse_values():
    """ Check input file validity """
    csv_file = csv.reader(open(sys.argv[1], 'r'))
    csv_file.__next__()  # Skip header
    line_count = 1

    for line in csv_file:
        if not line:  # Skip empty lines
            print("Empty line, skipping...")
            continue
        line_count = line_count + 1
        vdate = line[0]
        vtime = line[1]
        weight = line[2]
        height = line[3]
        bp_sys = line[4]
        bp_dia = line[5]
        heart_rate = line[6]
        # osat = line[7]
        # glucose = line[8]
        # steps = line[9]
        # calories = line[10]
        # sleep = line[11]

        event_id = str(uuid4())

        if vdate:
            if not vtime:   # Set the time to 00:00 when empty
                vtime = "00:00"
            isodate = validate_date(vdate, vtime)

            if weight:
                if height:
                    height = int(height)
                set_weight({'timestamp': isodate,
                            'event_id': event_id,
                            'height': height,
                            'weight': float(weight)})

            # At this point, we'll record the BP when both
            # systolic and diastolic readings are present
            if bp_sys and bp_dia:
                set_bp({'timestamp': isodate,
                        'event_id': event_id,
                        'systolic': int(bp_sys),
                        'diastolic': int(bp_dia)})

            if heart_rate:
                set_hr({'timestamp': isodate,
                        'event_id': event_id,
                        'heart_rate': int(heart_rate)})
        else:
            print("Error in date: Skipping line ", line_count)


if len(sys.argv) < 2:
    sys.exit("You need to specify a CSV file with the results")

print("Checking integrity of the datafile ...")
parse_values()



#!/usr/bin/env python3
##############################################################################
#
#    MyGNUHealth : Mobile and Desktop PHR node for GNU Health
#
#           MyGNUHealth is part of the GNU Health project
#
##############################################################################
#
#    GNU Health: The Libre Digital Health Ecosystem
#    Copyright (C) 2008-2024 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2024 GNU Solidario <health@gnusolidario.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import datetime
import os
import io
import sys
import math
import pygal
from pygal.style import Style
from tinydb import TinyDB, where
from mygnuhealth.myghconf import dbfile, bolfile
from mygnuhealth.lang import tr

# Set custom style for pygal charts

mygh_style = Style(
    background='transparent',
    plot_background='transparent',
    stroke_width='3',)


''' Import main and bol databases '''
maindb = TinyDB(dbfile)
boldb = TinyDB(bolfile)


def scale_xaxis(series_length, x_values):
    """ The method produces an array based on the original x_values
        of 15 elements or less, interspaced by the length of the array
        over MAX_LABELS
    """
    MAX_LABELS = 15

    index = 1
    counter = 0
    step = math.ceil(series_length / MAX_LABELS)

    while (step > 1 and index < series_length):
        if counter < step:
            x_values[index] = None
            counter = counter + 1
        else:
            counter = 0

        index = index + 1

    return x_values


def line_plot(title, y_legend, series, x_values=None, renderfmt=None):
    """ Method to plot basic charts using pygal, encapsulating
        the image object in memory
    """
    line_chart = pygal.Line(
        style=mygh_style,
        x_label_rotation=35,
        allow_interruptions=True)

    line_chart.title = title
    for title, data in series.items():
        line_chart.add(title, data)

    line_chart.x_labels = scale_xaxis(len(x_values), x_values)

    chart_io = io.BytesIO()
    if (renderfmt == 'svg'):
        line_chart.render(chart_io)  # Render to raw
        # Testing the svg format on a file
        # line_chart.render_to_file(filename='/tmp/foo.svg')
    else:
        line_chart.render_to_png(chart_io)  # PNG rendering

    chart_io.seek(0)
    return chart_io


def stacked_plot(title, series, x_values=None, renderfmt=None):

    stacked_chart = pygal.StackedBar()
    stacked_chart.title = title
    for title, data in series.items():
        stacked_chart.add(title, data)

    stacked_chart.x_labels = x_values

    chart_io = io.BytesIO()
    if (renderfmt == 'svg'):
        stacked_chart.render(chart_io)  # Render to raw
    else:
        stacked_chart.render_to_png(chart_io)  # Render to PNG

    chart_io.seek(0)
    return chart_io


def filter_by_date(data, date_start, date_end):
    start = datetime.datetime.fromisoformat(date_start).date()
    end = datetime.datetime.fromisoformat(date_end).date()

    start_is_found = []
    end_is_found = []
    results = []
    date_min = None

    for data_line in data:
        date = datetime.datetime.fromisoformat(
            data_line['timestamp']).date()
        if date_min:
            date_min = min(date, date_min)
        else:
            date_min = date
        if date == start:
            start_is_found.append('yes')
        if date == end:
            end_is_found.append('yes')
        if date >= start and date <= end:
            results.append(data_line)

    if date_min is not None:
        if start > date_min and 'yes' not in start_is_found:
            results.append({'timestamp': start.isoformat()})
    else:
        if 'yes' not in start_is_found:
            results.append({'timestamp': start.isoformat()})

    if 'yes' not in end_is_found:
        results.append({'timestamp': end.isoformat()})

    return results


def default_date_range(data=None):

    sorted_list = sorted(data, key=lambda sk: sk.get('timestamp'))

    if sorted_list:
        date_start = datetime.datetime.fromisoformat(
            sorted_list[0].get('timestamp')).strftime("%Y-%m-%d")
    else:
        date_start = (datetime.datetime.now() -
                      datetime.timedelta(days=7)).strftime("%Y-%m-%d")

    date_end = datetime.datetime.now().strftime("%Y-%m-%d")

    return (date_start, date_end)


def get_arch():
    if 'ANDROID_ARGUMENT' in os.environ:
        return "android"
    arch = sys.platform
    print(f"Running on {arch}")
    return arch


def pkg_path():
    if get_arch() == 'android':
        pass
        # return "/data/data/org.test.mygnuhealth/files/app/"
    else:
        moddir = os.path.dirname(os.path.abspath(__file__))
        return moddir


def get_user_profile(db):
    """Retrieves the user profile (DoB, sex, height ...)"""

    profile_table = db.table('profile')
    profile = None
    # Credentials table holds a singleton, so only one record
    if (len(profile_table) > 0):
        profile = profile_table.all()[0]
    return profile


def get_personal_key(db):
    """Retrieves the user personal key"""

    credentials_table = db.table('credentials')
    # Credentials table holds a singleton, so only one record
    personal_key = credentials_table.all()[0]['personal_key']
    return personal_key.encode()


def get_federation_account():
    """Retrieves the user GH Federation account, if any."""

    federation_table = maindb.table('federation')
    fedacct = None
    if (len(federation_table) > 0):
        # Federation Account table holds a singleton, so only one record
        res = federation_table.all()[0]
        if 'federation_account' in res.keys():
            fedacct = res['federation_account']
    return fedacct


def get_device(model):
    """Retrieves the device features and ID based on the model"""

    device_table = maindb.table('devices')
    # device = None

    res = device_table.search(where('model') == model)

    return res


def check_date(date):
    """ Verifies that the entered date is valid"""
    year, month, day = date
    try:
        datetime.date(int(year), int(month), int(day))
        return True
    except ValueError:
        print("Invalid date")
        return False


# moddir = pkg_path()
''' Import the Natural variants database from the data
directory, relative to the this mygnuhealth module
The datafile is loaded into the TinyDB "vardb" variable
'''

app_path = os.path.dirname(os.path.abspath(__file__))
varfile = os.path.join(app_path, 'data/variants.db')
vardb = TinyDB(varfile, access_mode='r')


class CodeNameMap():
    history = []
    mapinfo = None
    mapfunc = None
    current_mapinfo = []

    def __init__(self, mapinfo=None, mapfunc=None):
        self.mapinfo = mapinfo
        self.mapfunc = mapfunc

    def get_mapinfo(self):
        result = None
        if self.mapfunc:
            result = self.mapfunc()
        else:
            mapinfo = self.mapinfo
            result = [{'code': item['code'], 'name': tr._(item['name'])}
                      for item in mapinfo]

        self.current_mapinfo = result
        return result

    def get_names(self):
        mapinfo = self.get_mapinfo()
        return [item['name'] for item in mapinfo]

    def get_name(self, code):
        mapinfo = self.get_mapinfo()
        name = None

        for item in mapinfo:
            if item.get('code') == code:
                name = item.get('name')
                break

        return name

    def get_code(self, name):
        code = None
        mapinfo = self.get_mapinfo()

        for item in mapinfo:
            if item.get('name') == name:
                code = item.get('code')
                break

        if code is not None:
            history = list(reversed(self.history))
            for his in history:
                for item in his.get('mapinfo'):
                    if item.get('name') == name:
                        code = item.get('code')
                        break

        return code

    def update_history(self, name):
        self.history.append(
            {'name': name,
             'mapinfo': self.current_mapinfo})


class PageOfLife():
    """
    Page of Life
    The basic shema of PoL from GH Federation HIS, used  by Thalamus

    Attributes
    ----------
        boldb: TinyDB instance
            The book of life DB. It contains all the Pages of Life created
            by the user.

        pol_model : dict
            Dictionary holding the schema of the GNU Healtth Federation
            Health Information System database

        medical_context: In a page of life, when the medical domain is chosen,
            the user can choose

        social_context: The different contexts within the Social domain.

        Methods:
        --------
            create_pol: Creates a Page of Life associated the event / reading

    """

    boldb = TinyDB(bolfile)

    pol_model = dict.fromkeys([
        'book', 'page_date', 'age', 'domain', 'relevance', 'privacy',
        'context', 'measurements', 'genetic_info', 'summary', 'info',
        'node', 'author', 'author_acct', 'fsynced'
    ])

    def create_pol(self, pol_vals):
        """Creates a Page of Life associated to the reading

        Parameters
        ----------
        pol_vals: Takes all the values from the page of life, which is a
        dictionary. Some of them:
            domain: the domain (medical, psycho, social)
            context: the context within a domain (possible contexts are listed
                in the core module.
            genetic_info: variant, rsref, protein, gene, aa_change
            measurements: blood pressure, Osat, temp, heart & resp frequency,..
            summary: Short description / title of the page
            info: Extended information related to this page of life.
        """

        node = "mygnuhealth"  # The node name is generic. "mygnuhealth"
        fed_acct = get_federation_account()
        poltable = self.boldb.table('pol')
        page_of_life = self.pol_model.copy()

        domain = pol_vals.get('domain')
        context = pol_vals.get('context')

        if (fed_acct):
            #  If the Federation account does not exist, it will be
            #  a private entry, not linked to a book or author
            #  and it won't be shared in the GNU Health Federation

            print("Retrieved Federation Account: ", fed_acct)
            page_of_life['book'] = fed_acct
            page_of_life['author'] = fed_acct
            page_of_life['author_acct'] = fed_acct

        page_of_life['node'] = node
        page_of_life['page'] = pol_vals.get('page')
        page_of_life['page_date'] = pol_vals.get('page_date')
        page_of_life['domain'] = domain
        page_of_life['context'] = context
        page_of_life['relevance'] = pol_vals.get('relevance')
        page_of_life['privacy'] = pol_vals.get('privacy')
        page_of_life['genetic_info'] = pol_vals.get('genetic_info', '')
        page_of_life['measurements'] = pol_vals.get('measurements')
        page_of_life['summary'] = pol_vals.get('summary', '')
        page_of_life['info'] = pol_vals.get('info', '')

        # The fsync key reflects whether the page has been sent to the
        # GNU Health Federation HIS (Health Information System)
        page_of_life['fsynced'] = False
        # create the new PoL entry

        print("New Page of Life:", page_of_life.get('page'))
        data = page_of_life
        poltable.insert(data)

        # Sample measurements keys accepted by Thalamus / GH Federation HIS
        #  {'bp': {'systolic': 116, 'diastolic': 79}, 't': 36.0, 'hr': 756, '
        #    rr': 16,
        #  'osat': 99, 'wt': 68.0, 'ht': 168.0, 'bmi': 24.09, 'bg': 116}


class BookOfLifeFormatter():
    formatters = {}

    def add(self, name, formatter):
        self.formatters[name] = formatter

    def handle(self, name, args):
        if args:
            formatter = self.formatters.get(name)

            if formatter:
                return formatter(args)
            else:
                return args
        else:
            return ''


bol_domain_formatter = BookOfLifeFormatter()
bol_measurement_formatter = BookOfLifeFormatter()
bol_genetic_info_formatter = BookOfLifeFormatter()



#!/usr/bin/python

###########################################################
# date_picker.py is a fork of KivyCalendar:
#
# * KivyCalendar *
# URL: https://bitbucket.org/xxblx/kivycalendar
#      https://pypi.org/project/KivyCalendar/
# License: MIT License (MIT License)
# Author: Oleg Kozlov (xxblx)
###########################################################
import re

from calendar import Calendar, monthrange
from datetime import datetime
from dateutil.relativedelta import relativedelta

from kivy.uix.screenmanager import ScreenManager, Screen
from kivy.uix.popup import Popup
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.button import Button
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.textinput import TextInput
from kivy.uix.label import Label
from kivy.core.window import Window
from kivy.properties import (NumericProperty, BooleanProperty,
                             ObjectProperty, ReferenceListProperty)

from mygnuhealth.lang import tr


class DateRange(BoxLayout):
    touch_switch = BooleanProperty()
    popup_hint_x = NumericProperty(0.7)
    popup_hint_y = NumericProperty(0.7)
    popup_hint = ReferenceListProperty(popup_hint_x, popup_hint_y)
    date_start = ObjectProperty()
    date_end = ObjectProperty()
    update_range_function = ObjectProperty()

    def update_range(self, date_start, date_end, update_range_function):
        if self.check_date_range(date_start, date_end):
            update_range_function(date_start, date_end)
        else:
            popup = Popup(
                title=tr._('Wrong Date Range.'),
                content=Label(
                    text=tr._("Please check start and end of date.")),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

    def check_date_range(self, date_start, date_end):
        try:
            date_start = datetime.fromisoformat(date_start)
            date_end = datetime.fromisoformat(date_end)
            return date_start <= date_end
        except BaseException:
            return False


class DatePicker(TextInput):
    """Date picker is a textinput, if it focused shows popup with
    calendar which allows you to define the popup dimensions using
    popup_hint_x, popup_hint_y, and the popup_hint lists, for example
    in kv:

    DatePicker:
        popup_hint: 0.7,0.4

    would result in a size_hint of 0.7,0.4 being used to create the
    popup

    """
    popup_hint_x = NumericProperty(0.7)
    popup_hint_y = NumericProperty(0.7)
    popup_hint = ReferenceListProperty(popup_hint_x, popup_hint_y)
    init_date = ObjectProperty()
    touch_switch = BooleanProperty()

    def on_init_date(self, instance, value):
        self.init_ui()

    def on_touch_switch(self, instance, value):
        self.init_ui()

    def init_ui(self):

        self.text = get_init_date(self.init_date)

        # Popup Content.
        self.popup_content = DatePickerPopupContent(
            init_date=self.init_date,
            touch_switch=self.touch_switch,
            quit_function=self.quit_popup)

        # Popup
        self.popup = Popup(
            title="",
            content=self.popup_content,
            on_dismiss=self.update_value)

        self.bind(focus=self.show_popup)

    def quit_popup(self):
        self.popup.dismiss()

    def show_popup(self, isnt, val):
        """
        Open popup if textinput focused,
        and regardless update the popup size_hint
        """
        self.popup.size_hint = self.popup_hint
        if val:
            # Automatically dismiss the keyboard
            # that results from the textInput
            Window.release_all_keyboards()
            self.popup.open()

    def update_value(self, inst):
        """ Update textinput value on popup close """

        self.text = format_date(
            date=self.popup_content.get_active_date())
        self.popup_content.clear_text_input()

        self.focus = False


class DatePickerPopupContent(BoxLayout):
    init_date = ObjectProperty()
    touch_switch = ObjectProperty()
    calendar = ObjectProperty()
    date_input = ObjectProperty()
    quit_function = ObjectProperty()

    def get_active_date(self):
        return self.calendar.active_date

    def quit_calendar(self):
        self.quit_function()
        self.clear_text_input()

    def clear_text_input(self):
        self.date_input.text = ''

    def select_date(self, text):
        try:
            if re.search(r'[+ymdw]', text):
                self._select_date_by_delta(text)
            else:
                self._select_date(text)
        except BaseException:
            pass

    def _select_date(self, text):

        def format_year(year, cur_year):
            year = str(year)
            cur_year = str(cur_year)
            if len(year) < 4:
                return int(cur_year[0:(4 - len(year))] + year)
            else:
                return int(year)

        def format_month(month, cur_month):
            if month > 12:
                return cur_month
            else:
                return month

        def format_day(day, cur_day):
            if day > 31:
                return cur_day
            else:
                return day

        date_now = datetime.now()
        cur_year = date_now.year
        cur_month = date_now.month
        cur_day = date_now.day
        year, month, day = cur_year, cur_month, cur_day

        # Support 'y-m-d' 'y/m/d' and 'y m d'.
        text = re.sub(r'[^0-9-/ ]', '', text)
        text = re.sub(r'[/ ]', '-', text)
        info = [int(x) for x in text.split('-') if x != '']

        if len(info) == 3:
            year = format_year(info[0], cur_year)
            month = format_month(info[1], cur_month)
            day = format_day(info[2], cur_day)
        elif len(info) == 2:
            month = format_month(info[0], cur_month)
            day = format_day(info[1], cur_day)
        elif len(info) == 1:
            day = format_day(info[0], cur_day)
        else:
            pass

        date_str = f'{year:04}-{month:02}-{day:02}'
        date = datetime.fromisoformat(date_str).isoformat()
        self.calendar.select_date(date)

    def _select_date_by_delta(self, text):

        date_now = datetime.now()
        years = months = weeks = days = 0

        rule_list = re.sub(r'[^0-9-+wmdy ]', '', text).split(' ')

        for rule in rule_list:
            if rule != '':
                categery = rule[-1]
                if rule[0] == '-':
                    sign = -1
                else:
                    sign = 1
            else:
                categery = ''
                sign = 1

            n = re.sub(r'[^0-9]', '', rule)

            if n != '':
                num = int(n) * sign
                if categery == 'y':
                    years = num
                elif categery == 'm':
                    months = num
                elif categery == 'w':
                    weeks = num
                else:
                    days = num

        date = date_now + relativedelta(
            years=years, months=months,
            weeks=weeks, days=days)

        self.calendar.select_date(date.strftime("%Y-%m-%d"))

    def select_today(self):
        date = datetime.now().strftime("%Y-%m-%d")
        self.calendar.select_date(date, quit=True)


class CalendarWidget(RelativeLayout):
    """ Basic calendar widget """
    init_date = ObjectProperty()
    touch_switch = ObjectProperty()
    quit_function = ObjectProperty()

    def __init__(self, *args, **kwargs):
        super(CalendarWidget, self).__init__(*args, **kwargs)
        self.prepare_data()
        self.init_ui()

    def select_date(self, date, quit=False):
        self.init_date = date
        self.refresh_ui()
        if quit:
            self.quit_calendar()

    def refresh_ui(self):
        self.clear_widgets()
        self.prepare_data()
        self.init_ui()

    def quit_calendar(self):
        if self.quit_function:
            self.quit_function()

    def init_ui(self):
        self.left_arrow = ArrowButton(
            text="<", on_press=self.go_prev,
            pos_hint={"top": 1, "left": 0})

        self.right_arrow = ArrowButton(
            text=">", on_press=self.go_next,
            pos_hint={"top": 1, "right": 1})

        self.add_widget(self.left_arrow)
        self.add_widget(self.right_arrow)

        # Title
        self.title_label = MonthYearLabel(text=self.title)
        self.add_widget(self.title_label)

        # ScreenManager
        self.sm = MonthsManager()
        self.add_widget(self.sm)

        self.create_month_scr(self.quarter[1], toogle_today=True)

    def create_month_scr(self, month, toogle_today=False):
        """ Screen with calendar for one month """

        scr = Screen()
        scr.name = format_date(
            year=self.active_date['year'],
            month=self.active_date['month'])

        # Grid for days
        grid_layout = ButtonsGrid()
        scr.add_widget(grid_layout)

        # Days abbrs
        for i in range(7):
            if i >= 5:  # weekends
                line = DayAbbrWeekendLabel(text=self.days_abrs[i])
            else:  # work days
                line = DayAbbrLabel(text=self.days_abrs[i])

            grid_layout.add_widget(line)

        # Buttons with days numbers
        for week in month:
            for day in week:
                if day[1] >= 5:  # weekends
                    tbtn = DayNumWeekendButton(text=str(day[0]))
                else:  # work days
                    tbtn = DayNumButton(text=str(day[0]))

                tbtn.bind(on_press=self.get_btn_value)

                if toogle_today:
                    # Down today button
                    if day[0] == self.active_date['day'] and day[2] == 1:
                        tbtn.state = "down"
                # Disable buttons with days from other months
                if day[2] == 0:
                    tbtn.disabled = True

                grid_layout.add_widget(tbtn)

        self.sm.add_widget(scr)

    def prepare_data(self):
        """ Prepare data for showing on widget loading """

        # Get days abbrs and month names lists
        self.days_abrs = get_days_abbrs()

        # Today date
        self.active_date = get_init_date_list(self.init_date)

        # Set title
        self.title = format_date(date=self.active_date)

        # Quarter where current month in the self.quarter[1]
        self.get_quarter()

    def get_quarter(self):
        """ Get caledar and months/years nums for quarter """

        self.quarter_nums = calc_quarter(
            self.active_date['year'],
            self.active_date['month'])

        self.quarter = get_quarter(
            self.active_date['year'],
            self.active_date['month'])

    def get_btn_value(self, inst):
        """ Get day value from pressed button """

        self.active_date['day'] = int(inst.text)
        self.quit_calendar()

    def go_prev(self, inst):
        """ Go to screen with previous month """

        # Change active date
        self.active_date = {
            'day': self.active_date['day'],
            'month': self.quarter_nums['previous']['month'],
            'year': self.quarter_nums['previous']['year']}

        # Name of prev screen
        prev_scr_name = format_date(
            year=self.quarter_nums['previous']['year'],
            month=self.quarter_nums['previous']['month'])

        # If it's doen't exitst, create it
        if not self.sm.has_screen(prev_scr_name):
            self.create_month_scr(self.quarter[0])

        self.sm.current = prev_scr_name
        self.sm.transition.direction = "left"

        self.get_quarter()
        self.title = format_date(
            year=self.active_date['year'],
            month=self.active_date['month'])

        self.title_label.text = self.title

    def go_next(self, inst):
        """ Go to screen with next month """

        # Change active date
        self.active_date = {
            'day': self.active_date['day'],
            'month': self.quarter_nums['next']['month'],
            'year': self.quarter_nums['next']['year']}

        # Name of prev screen
        next_scr_name = format_date(
            year=self.quarter_nums['next']['year'],
            month=self.quarter_nums['next']['month'])

        # If it's doen't exitst, create it
        if not self.sm.has_screen(next_scr_name):
            self.create_month_scr(self.quarter[2])

        self.sm.current = next_scr_name
        self.sm.transition.direction = "right"

        self.get_quarter()

        self.title = format_date(
            year=self.active_date['year'],
            month=self.active_date['month'])

        self.title_label.text = self.title

    def on_touch_move(self, touch):
        """ Switch months pages by touch move """

        if self.touch_switch:
            # Left - prev
            if touch.dpos[0] < -30:
                self.go_prev(None)
            # Right - next
            elif touch.dpos[0] > 30:
                self.go_next(None)


class ArrowButton(Button):
    pass


class MonthYearLabel(Label):
    pass


class MonthsManager(ScreenManager):
    pass


class ButtonsGrid(GridLayout):
    pass


class DayAbbrLabel(Label):
    pass


class DayAbbrWeekendLabel(DayAbbrLabel):
    pass


class DayButton(ToggleButton):
    pass


class DayNumButton(DayButton):
    pass


class DayNumWeekendButton(DayButton):
    pass


def format_date(date=None, year=None, month=None, day=None):
    if date:
        year = date['year']
        month = date['month']
        day = date['day']

    year = year and f'{year:04}' or ''
    month = month and f'-{month:02}' or ''
    day = day and f'-{day:02}' or ''

    return year + month + day


def get_days_abbrs():
    """ Return list with days abbreviations """

    return [tr._('Mon'), tr._('Tue'), tr._('Wed'),
            tr._('Thu'), tr._('Fri'), tr._('Sat'),
            tr._('Sun')]


def get_month(y, m):
    """
    Return list of month's weeks, which day
    is a turple (<month day number>, <weekday number>)
    """

    cal = Calendar(firstweekday=0)
    month = cal.monthdays2calendar(y, m)

    # Add additional num to every day which mark from
    # this or from other day that day numer
    for week in range(len(month)):
        for day in range(len(month[week])):
            _day = month[week][day]
            if _day[0] == 0:
                this = 0
            else:
                this = 1
            _day = (_day[0], _day[1], this)
            month[week][day] = _day

    # Days numbers of days from preious and next monthes
    # marked as 0 (zero), replace it with correct numbers
    # If month include 4 weeks it hasn't any zero
    if len(month) == 4:
        return month

    quater = calc_quarter(y, m)

    # Zeros in first week
    fcount = 0
    for i in month[0]:
        if i[0] == 0:
            fcount += 1

    # Zeros in last week
    lcount = 0
    for i in month[-1]:
        if i[0] == 0:
            lcount += 1

    if fcount:

        n = monthrange(
            quater['previous']['year'],
            quater['previous']['month'])[1]

        for i in range(fcount):
            month[0][i] = (n - (fcount - 1 - i), i, 0)

    if lcount:
        # First day of next month
        n = 1

        for i in range(lcount):
            month[-1][-lcount + i] = (n + i, 7 - lcount + i, 0)

    return month


def get_quarter(y, m):
    """ Get quarter where m is a middle month """

    result = []
    quarter = calc_quarter(y, m)
    for key in ['previous', 'current', 'next']:
        result.append(get_month(
            quarter[key]['year'], quarter[key]['month']))

    return result


def calc_quarter(y, m):
    """ Calculate previous and next month """

    # Previous / Next month's year number and month number
    prev_y = y
    prev_m = m - 1
    next_y = y
    next_m = m + 1

    if m == 1:
        prev_m = 12
        prev_y = y - 1
    elif m == 12:
        next_m = 1
        next_y = y + 1

    return {'previous': {'year': prev_y, 'month': prev_m},
            'current': {'year': y, 'month': m},
            'next': {'year': next_y, 'month': next_m}}


def get_init_date_list(init_date):
    """ Return dict with today date """

    try:
        date = datetime.fromisoformat(init_date)
        return {
            'day': date.day,
            'month': date.month,
            'year': date.year}
    except BaseException:
        return {
            'day': datetime.now().day,
            'month': datetime.now().month,
            'year': datetime.now().year}


def get_init_date(init_date):
    """ Return today date """
    try:
        datetime.fromisoformat(init_date)
        return init_date
    except BaseException:
        return datetime.now().strftime("%Y-%m-%d")



####################################################################
#   Copyright (C) 2020-2024 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2024 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import datetime
from tinydb import Query
from uuid import uuid4
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from mygnuhealth.lang import tr
from mygnuhealth.core import (check_date, PageOfLife,
                              vardb, CodeNameMap,
                              bol_domain_formatter,
                              bol_genetic_info_formatter)


class PoL():
    """This class creates a new page in the user's Book of Life

        Attributes:
        -----------
            wrongDate: Signal to emit when an invalid date is found
            todayDate: Property with current date

        Methods:
        --------
            get_domains: Returns main domains (medical, social, biographical..)

    """

    domain = CodeNameMap(
        mapinfo=[{'code': 'medical', 'name': tr.N_('Medical')},
                 {'code': 'social', 'name': tr.N_('Social')},
                 {'code': 'biographical', 'name': tr.N_('Biographical')},
                 {'code': 'lifestyle', 'name': tr.N_('Lifestyle')},
                 {'code': 'other', 'name': tr.N_('Other')}
                 ])

    medical_context = CodeNameMap(
        mapinfo=[
            {'code': 'health_condition', 'name': tr.N_('Health Condition')},
            {'code': 'encounter', 'name': tr.N_('Encounter')},
            {'code': 'procedure', 'name': tr.N_('Procedure')},
            {'code': 'self_monitoring', 'name': tr.N_('Self monitoring')},
            {'code': 'immunization', 'name': tr.N_('Immunization')},
            {'code': 'prescription', 'name': tr.N_('Prescription')},
            {'code': 'surgery', 'name': tr.N_('Surgery')},
            {'code': 'hospitalization', 'name': tr.N_('Hospitalization')},
            {'code': 'lab', 'name': tr.N_('Lab test')},
            {'code': 'dx_imaging', 'name': tr.N_('Dx Imaging')},
            {'code': 'genetics', 'name': tr.N_('Genetics')},
            {'code': 'family', 'name': tr.N_('Family history')},
        ])

    social_context = CodeNameMap(
        mapinfo=[
            {'code': 'social_gradient',
             'name': tr.N_('Social Gradient / Equity')},
            {'code': 'stress', 'name': tr.N_('Stress')},
            {'code': 'early_life_development',
             'name': tr.N_('Early life development')},
            {'code': 'social_exclusion',
             'name': tr.N_('Social exclusion')},
            {'code': 'working_conditions',
             'name': tr.N_('Working conditions')},
            {'code': 'education', 'name': tr.N_('Education')},
            {'code': 'physical_environment',
             'name': tr.N_('Physical environment')},
            {'code': 'unemployment', 'name': tr.N_('Unemployment')},
            {'code': 'social_support', 'name': tr.N_('Social Support')},
            {'code': 'addiction', 'name': tr.N_('Addiction')},
            {'code': 'food', 'name': tr.N_('Food')},
            {'code': 'transportation', 'name': tr.N_('Transportation')},
            {'code': 'health_services', 'name': tr.N_('Health services')},
            {'code': 'family_functionality',
             'name': tr.N_('Family functionality')},
            {'code': 'family_violence', 'name': tr.N_('Family violence')},
            {'code': 'bullying', 'name': tr.N_('Bullying')},
            {'code': 'war', 'name': tr.N_('War')},
            {'code': 'misc', 'name': tr.N_('Misc')},
        ])

    lifestyle_context = CodeNameMap(
        mapinfo=[
            {'code': 'physical_activity', 'name': tr.N_('Physical Activity')},
            {'code': 'nutrition', 'name': tr.N_('Nutrition')},
            {'code': 'sleep', 'name': tr.N_('Sleep')},
            {'code': 'social_activities', 'name': tr.N_('Social Activities')},
        ])

    biographical_context = CodeNameMap(
        mapinfo=[
            {'code': 'birth', 'name': tr.N_('Birth')},
            {'code': 'death', 'name': tr.N_('Death')},
            {'code': 'misc', 'name': tr.N_('Misc')}
        ])

    other_context = CodeNameMap(
        mapinfo=[
            {'code': 'misc', 'name': tr.N_('Misc')},
        ])

    relevance = CodeNameMap(
        mapinfo=[
            {'code': 'normal', 'name': tr.N_('Normal')},
            {'code': 'important', 'name': tr.N_('Important')},
            {'code': 'critical', 'name': tr.N_('Critical')},
        ])

    @classmethod
    def get_domain_names(cls):
        return cls.domain.get_names()

    @classmethod
    def get_domain_name(cls, domain_code):
        return cls.domain.get_name(domain_code)

    @classmethod
    def get_domain_code(cls, domain_name):
        return cls.domain.get_code(domain_name)

    @classmethod
    def get_context_name(cls, context_code):
        context_name = None
        for ctx in [cls.social_context,
                    cls.medical_context,
                    cls.lifestyle_context,
                    cls.biographical_context,
                    cls.other_context]:
            context_name = ctx.get_name(context_code)
            if context_name:
                break

        return context_name

    @classmethod
    def get_context_names(cls, domain_name):
        domain_code = cls.get_domain_code(domain_name)
        if domain_code == 'social':
            domain_contexts = cls.social_context
        if domain_code == 'medical':
            domain_contexts = cls.medical_context
        if domain_code == 'lifestyle':
            domain_contexts = cls.lifestyle_context
        if domain_code == 'biographical':
            domain_contexts = cls.biographical_context
        if domain_code == 'other':
            domain_contexts = cls.other_context

        return domain_contexts.get_names()

    @classmethod
    def update_context_names(cls, domain_name):
        """ Set the value of the domain from the selection"""
        domain_code = cls.get_domain_code(domain_name)

        if domain_code == 'social':
            pol_context = cls.social_context
        if domain_code == 'medical':
            pol_context = cls.medical_context
        if domain_code == 'lifestyle':
            pol_context = cls.lifestyle_context
        if domain_code == 'biographical':
            pol_context = cls.biographical_context
        if domain_code == 'other':
            pol_context = cls.other_context

        return pol_context.get_names()

    @classmethod
    def domain_context_is_genetic(cls, context_name):
        context_code = cls.get_context_code(context_name)
        return context_code == 'genetics'

    @classmethod
    def get_context_code(cls, context_name):
        context_code = None
        # Should we add domain argument to let method faster?
        for ctx in [cls.social_context,
                    cls.medical_context,
                    cls.lifestyle_context,
                    cls.biographical_context,
                    cls.other_context]:
            context_code = ctx.get_code(context_name)
            if context_code:
                break

        return context_code

    @classmethod
    def get_relevance_names(cls):
        return cls.relevance.get_names()

    @classmethod
    def get_relevance_code(cls, relevance_name):
        return cls.relevance.get_code(relevance_name)

    @classmethod
    def bol_domain_formatter(cls, args):
        domain_code, context_code = args[0], args[1]
        return '{0} ({1})'.format(
            cls.get_domain_name(domain_code) or '',
            cls.get_context_name(context_code) or '')

    def get_rsinfo(self):
        return self.rsinfo

    def get_date():
        """
        Returns the date packed into an array (day,month,year, hour, min)
        """
        rightnow = datetime.datetime.now()
        dateobj = []
        dateobj.append(str(rightnow.day))
        dateobj.append(str(rightnow.month))
        dateobj.append(str(rightnow.year))
        dateobj.append(str(rightnow.hour))
        dateobj.append(str(rightnow.minute))
        return dateobj

    def new_page(data):
        page_id = str(uuid4())

        pol_vals = {
            'page': page_id,
            'page_date': data['page_date'],
            'domain': data['domain'],
            'context': data['context'],
            'relevance': data['relevance'],
            'privacy': data['privacy'],
            'summary': data['summary'],
            'info': data['info']
        }
        if (data['context'] == 'genetics'):
            pol_vals.update({'genetic_info': data['genetic_info']})

        PageOfLife.create_pol(PageOfLife, pol_vals)

    def checkSNP(rs):
        rsinfo = {}
        if rs:
            Rsnp = Query()
            res = vardb.search(Rsnp.dbsnp == rs)
            if len(res) > 0:
                res = res[0]
                rsinfo = {
                    'rsid': res['dbsnp'],
                    'gene': res['gene'],
                    'aa_change': res['aa_change'],
                    'variant': res['variant'],
                    'protein': res['protein'],
                    'category': res['category'],
                    'disease': res['disease']
                }

                print(rsinfo)
            else:
                print(f"{rs} not found")

        return (rsinfo)

    @classmethod
    def createPage(cls, page_date, domain, context, relevance, private_page,
                   genetic_info, summary, info, use_name=False):
        # Retrieves the information from the initialization form
        # Creates the page from the information on the form
        if use_name:
            domain_code = cls.get_domain_code(domain)
            context_code = cls.get_context_code(context)
            relevance_code = cls.get_relevance_code(relevance)
        else:
            domain_code = domain
            context_code = context
            relevance_code = relevance

        if (page_date):
            if (check_date(page_date[:3])):
                # Sets the page of life date and time
                year, month, day, hour, minute = page_date

                daterp = datetime.datetime(
                    int(year), int(month),
                    int(day), int(hour),
                    int(minute)).isoformat()

                page = {'page_date': daterp,
                        'domain': domain_code,
                        'context': context_code,
                        'relevance': relevance_code,
                        'privacy': private_page,
                        'summary': summary,
                        'info': info
                        }
                if (context_code == 'genetics'):
                    rsinfo = {
                        'rsid': genetic_info[0],
                        'gene': genetic_info[1],
                        'aa_change': genetic_info[2],
                        'variant': genetic_info[3],
                        'protein': genetic_info[4],
                        'significance': genetic_info[5],
                        'disease': genetic_info[6]
                    }
                    page.update({'genetic_info': rsinfo})
                PoL.new_page(page)
                popup = Popup(
                    title=tr._('Success'),
                    content=Label(
                        text=tr._("Page of Life "
                                  "successfully created!")),
                    size_hint=(0.5, 0.5), auto_dismiss=True)
                popup.open()
                return True

            else:
                popup = Popup(
                    title=tr._('Error'),
                    content=Label(
                        text=tr._("Wrong date")),
                    size_hint=(0.5, 0.5), auto_dismiss=True)
                popup.open()

    @classmethod
    def bol_genetic_info_formatter(cls, args):
        rsid = str(args.get('rsid')) or ''
        gene = str(args.get('gene')) or ''
        aa_change = str(args.get('aa_change')) or ''
        variant = str(args.get('variant')) or ''
        protein = str(args.get('protein')) or ''
        significance = str(args.get('significance')) or ''
        disease = str(args.get('disease')) or ''

        return (tr._("RefSNP ID: ") + rsid + '\n' +
                tr._("Gene: ") + gene + '\n' +
                tr._("AA_change: ") + aa_change + '\n' +
                tr._("Natural variant: ") + variant + '\n' +
                tr._("Protein ID: ") + protein + '\n' +
                tr._("Significance: ") + significance + '\n' +
                tr._("Disease: ") + disease)


bol_genetic_info_formatter.add(
    'genetic_info', PoL.bol_genetic_info_formatter)


bol_domain_formatter.add(
    'domain', PoL.bol_domain_formatter)



#!/usr/bin/env python3
##############################################################################
#
#    MyGNUHealth : Mobile and Desktop PHR node for GNU Health
#
#           MyGNUHealth is part of the GNU Health project
#
##############################################################################
#
#    GNU Health: The Libre Digital Health Ecosystem
#    Copyright (C) 2008-2024 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2024 GNU Solidario <health@gnusolidario.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import os
import gettext
from kivy.lang import Observable


class Lang(Observable):
    observers = []
    lang = None

    def __init__(self, defaultlang):
        super(Lang, self).__init__()
        self.ugettext = lambda text: text
        self.lang = defaultlang
        self.switch_language(self.lang)

    def _(self, text):
        if isinstance(text, str) and text != '':
            return self.ugettext(text)
        else:
            return text

    # Used to deferred translations.
    def N_(self, text):
        return text

    def fbind(self, name, func, args, **kwargs):
        if name == "_":
            self.observers.append((func, args, kwargs))
        else:
            return super(Lang, self).fbind(name, func, *args, **kwargs)

    def funbind(self, name, func, args, **kwargs):
        if name == "_":
            key = (func, args, kwargs)
            if key in self.observers:
                self.observers.remove(key)
        else:
            return super(Lang, self).funbind(name, func, *args, **kwargs)

    def switch_language(self, lang):
        try:
            locale_dir = os.path.join(
                os.path.dirname(__file__), 'data', 'locale')
            locales = gettext.translation(
                'mygnuhealth', locale_dir, languages=[lang])
            self.ugettext = locales.gettext
            self.lang = lang

            # update all the kv rules attached to this text
            for func, largs, kwargs in self.observers:
                func(largs, None, None)
        except BaseException:
            print('Warn: switch_language can not ' +
                  f'find locale file of "{lang}"!!!')


tr = Lang("en")



from kivy.uix.screenmanager import Screen
import datetime
from uuid import uuid4
from kivy.properties import ObjectProperty
from kivy.core.image import Image as CoreImage
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from mygnuhealth.core import (maindb, PageOfLife, line_plot,
                              filter_by_date, default_date_range,
                              bol_measurement_formatter)
from mygnuhealth.lang import tr


class TrackerBioGlucoseScreen(Screen):
    """Class that manages the person blood glucose readings

        Attributes:
        -----------

        Methods:
        --------
            set_values: Places the new reading values on the 'glucose'
            table and creates the associated page of life
            read_glucose: Retrieve the blood glucose levels history
            getGlucose: Extracts the latest readings from Glucose
    """

    def read_glucose():
        # Retrieve the blood glucose levels history
        glucose = maindb.table('glucose')
        glucosehist = glucose.all()
        return glucosehist

    def getGlucose():
        # Extracts the latest readings from Glucose
        glucosehist = TrackerBioGlucoseScreen.read_glucose()
        glucoseobj = ['', '']
        if (glucosehist):  # Enter this block if there is a history
            glucose = glucosehist[-1]  # Get the latest (newest) record
            dateobj = datetime.datetime.fromisoformat(glucose['timestamp'])
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y - %H:%M"))

            glucoseobj = [str(date_repr), str(glucose['glucose'])]
        return glucoseobj

    def validate_values(self, glucose):
        # Check for sanity on values before saving them
        rc = 0
        errors = []

        if glucose:
            if (int(glucose) in range(20, 800)):
                glucose = int(glucose)
            else:
                rc = -1
                errors.append("Glucose")
        else:
            glucose = 0

        if (rc == 0):
            self.set_values(glucose)

        else:
            popup = Popup(
                title=tr._('Wrong values'),
                content=Label(
                    text=tr._("Please check {0}").format(errors)),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

    def set_values(self, blood_glucose):
        """Places the new reading values on the 'glucose' table

        Parameters
        ----------
        blood_glucose: value coming from the getvals method
        """

        glucose = maindb.table('glucose')
        current_date = datetime.datetime.now().isoformat()
        domain = 'medical'
        context = 'self_monitoring'

        if blood_glucose > 0:
            event_id = str(uuid4())
            synced = False
            monitor_vals = {'timestamp': current_date,
                            'event_id': event_id,
                            'synced': synced,
                            'glucose': blood_glucose
                            }
            glucose.insert(monitor_vals)

            print("Saved glucose", event_id, synced, blood_glucose,
                  current_date)

            # Create a new PoL with the values
            # within the medical domain and the self monitoring context
            pol_vals = {
                'page': event_id,
                'page_date': current_date,
                'domain': domain,
                'context': context,
                'measurements': [{'bg': blood_glucose}]
            }

            # Create the Page of Life associated to this blood glucose reading
            PageOfLife.create_pol(PageOfLife, pol_vals)

    @classmethod
    def bol_measurement_formatter_bg(cls, args):
        bg = str(args) or ''
        return tr._("Blood Glucose: ") + bg + ' ' + tr._("mg/dl")


bol_measurement_formatter.add(
    'bg',
    TrackerBioGlucoseScreen.bol_measurement_formatter_bg)


class TrackerBioGlucoseStatsScreen(Screen):
    glucose_plot = ObjectProperty()
    date_start = ObjectProperty()
    date_end = ObjectProperty()

    def on_pre_enter(self):
        date_range = default_date_range(
            data=TrackerBioGlucoseScreen.read_glucose()
        )
        self.date_start = date_range[0]
        self.date_end = date_range[1]
        # Update / Refresh the chart anytime we access the stats screen
        self.glucose_plot = self.Glucoseplot()

    def update_pots(self, date_start, date_end):
        self.date_start = date_start
        self.date_end = date_end
        self.glucose_plot = self.Glucoseplot()

    # Plotting - Glycemia
    def Glucoseplot(self):
        # Retrieves the history and packages into an array.
        glucosehist = filter_by_date(
            TrackerBioGlucoseScreen.read_glucose(),
            self.date_start,
            self.date_end)

        glucose = []
        glucose_date = []
        sorted_list = sorted(glucosehist, key=lambda sk: sk.get('timestamp'))

        for element in sorted_list:
            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            glucose_date.append(dateobj.strftime('%Y-%m-%d'))
            glucose.append(element.get('glucose'))

        series_glucose = {tr._('Blood Glucose Level'): glucose}

        chart_io = line_plot(
            title=tr._('Blood Glucose Level') + ' (' + tr._('mg/dl') + ')',
            series=series_glucose,
            y_legend=tr._('mg/dl'),
            x_values=glucose_date,
            renderfmt='png')

        return CoreImage(chart_io, ext="png").texture



from kivy.uix.screenmanager import Screen
import datetime
from uuid import uuid4
from kivy.properties import ObjectProperty
from kivy.core.image import Image as CoreImage
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from mygnuhealth.core import (maindb, PageOfLife, line_plot,
                              filter_by_date, default_date_range,
                              stacked_plot, bol_measurement_formatter)
from mygnuhealth.lang import tr


class TrackerLifestylePactivityScreen(Screen):
    """ Class that manages the person Physical Activity related readings
        We set and retrieve the latest values from the physical activity
        subsystem (aerobic, anaerobic and steps)

        Attributes:
        -----------

        Methods:
        --------
            set_values: Places the new reading values on the physical
                        activity and creates the associated page of life

    """

    # PHYSICAL ACTIVITY
    def read_pa():
        # Retrieve the physical activity history
        pactivity = maindb.table('physicalactivity')
        pahist = pactivity.all()
        return pahist

    def getPA():
        # Extracts the latest readings from Physical Activity table
        pahist = TrackerLifestylePactivityScreen.read_pa()
        paobj = ['', '', '', '']  # Init to empty string to avoid undefined val
        if pahist:
            pa = pahist[-1]  # Get the latest (newest) record

            dateobj = datetime.datetime.fromisoformat(pa['timestamp'])
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y - %H:%M"))

            paobj = [str(date_repr), str(pa['aerobic']),
                     str(pa['anaerobic']), str(pa['steps'])]

        return paobj

    def validate_values(self, aerobic, anaerobic, steps):
        # Check for sanity on values before saving them
        rc = 0
        errors = []
        if aerobic:
            if (int(aerobic) in range(1, 1400)):
                aerobic = int(aerobic)
            else:
                rc = -1
                errors.append("Aerobic")
        else:
            aerobic = 0

        if anaerobic:
            if (int(anaerobic) in range(1, 1400)):
                anaerobic = int(anaerobic)
            else:
                errors.append("Anaerobic")
                rc = -1
        else:
            anaerobic = 0

        if steps:
            if (int(steps) in range(1, 200000)):
                steps = int(steps)
            else:
                rc = -1
                errors.append("Steps")
        else:
            steps = 0

        if (rc == 0):
            self.set_values(aerobic, anaerobic, steps)

        else:
            popup = Popup(
                title=tr._('Wrong values'),
                content=Label(
                    text=tr._("Please check {0}").format(errors)),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

    def set_values(self, aerobic, anaerobic, steps):
        pactivity = maindb.table('physicalactivity')
        current_date = datetime.datetime.now().isoformat()
        domain = 'lifestyle'
        context = 'physical_activity'

        pa_event_id = str(uuid4())
        synced = False
        pactivity.insert({'timestamp': current_date,
                          'event_id': pa_event_id,
                          'synced': synced,
                          'aerobic': aerobic,
                          'anaerobic': anaerobic,
                          'steps': steps})

        print("Saved Physical Activity", pa_event_id, synced, aerobic,
              anaerobic, steps, current_date)

        # Page of Life block related to Physical Activity
        event_id = str(uuid4())
        monitor_readings = [
            {'pa': {'aerobic': aerobic,
                    'anaerobic': anaerobic,
                    'steps': steps}},
        ]

        pol_vals = {
            'page': event_id,
            'page_date': current_date,
            'domain': domain,
            'context': context,
            'measurements': monitor_readings
        }

        # Create the Page of Life associated to this reading
        PageOfLife.create_pol(PageOfLife, pol_vals)

    @classmethod
    def bol_measurement_formatter_pa(cls, args):
        aerobic = str(args.get('aerobic')) or ''
        anaerobic = str(args.get('anaerobic')) or ''
        steps = str(args.get('steps')) or ''

        return (tr._("Aerobic: ") + aerobic + '\n' +
                tr._("Anaerobic: ") + anaerobic + '\n' +
                tr._("Steps: ") + steps)


bol_measurement_formatter.add(
    'pa', TrackerLifestylePactivityScreen.bol_measurement_formatter_pa)


class TrackerLifestylePactivityStatsScreen(Screen):
    date_start = ObjectProperty()
    date_end = ObjectProperty()
    pa_plot = ObjectProperty()
    steps_plot = ObjectProperty()

    def on_pre_enter(self):
        date_range = default_date_range(
            data=TrackerLifestylePactivityScreen.read_pa()
        )
        self.date_start = date_range[0]
        self.date_end = date_range[1]
        # Update / Refresh the chart anytime we access the stats screen
        self.pa_plot, self.steps_plot = self.PAplot()

    def update_pots(self, date_start, date_end):
        self.date_start = date_start
        self.date_end = date_end
        self.pa_plot, self.steps_plot = self.PAplot()

    # Plotting - Physical Activity
    def PAplot(self):
        # Retrieves all the history and packages into an array.
        pahist = filter_by_date(
            TrackerLifestylePactivityScreen.read_pa(),
            self.date_start,
            self.date_end)

        pa_aerobic = []
        pa_anaerobic = []
        pa_steps = []
        pa_date = []

        # Sort the list of dictionaries using the timestamp as key
        sorted_list = sorted(pahist, key=lambda sk: sk.get('timestamp'))

        for element in sorted_list:
            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            pa_date.append(dateobj.strftime('%Y-%m-%d'))
            pa_aerobic.append(element.get('aerobic'))
            pa_anaerobic.append(element.get('anaerobic'))
            pa_steps.append(element.get('steps'))

        series_pa = {tr._('Aerobic'): pa_aerobic,
                     tr._('Anaerobic'): pa_anaerobic}
        series_steps = {tr._('Steps'): pa_steps}

        chart_pa = stacked_plot(
            tr._('Physical Activity') + ' (' + tr._('Minutes') + ')',
            series_pa, x_values=pa_date, renderfmt='png')

        chart_steps = line_plot(
            title=tr._("Steps"),
            series=series_steps,
            y_legend=tr._('Steps'),
            x_values=pa_date,
            renderfmt='png')

        return [CoreImage(chart_pa, ext="png").texture,
                CoreImage(chart_steps, ext="png").texture]



from kivy.uix.screenmanager import Screen
import datetime
from uuid import uuid4
from kivy.properties import ObjectProperty
from kivy.core.image import Image as CoreImage
from kivy.uix.label import Label
from kivy.uix.popup import Popup

from mygnuhealth.core import (maindb, PageOfLife, line_plot,
                              filter_by_date, default_date_range,
                              bol_measurement_formatter)
from mygnuhealth.lang import tr


class TrackerBioOsatScreen(Screen):
    """Class that manages the person Hb oxygen saturation readings

        Attributes:
        -----------

        Methods:
        --------
            set_values: Places the new reading values on the 'osat'
                        and creates the associated page of life
            read_osat: Retrieve the blood osat levels history
            getOsat: Extracts the latest readings from Osat
    """

    # OSAT
    def read_osat():
        # Retrieve the blood osat levels history
        osat = maindb.table('osat')
        osathist = osat.all()
        return osathist

    def getOsat():
        # Extracts the latest readings from Osat
        osathist = TrackerBioOsatScreen.read_osat()
        osatobj = ['', '']
        if (osathist):
            osat = osathist[-1]  # Get the latest (newest) record
            dateobj = datetime.datetime.fromisoformat(osat['timestamp'])
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y - %H:%M"))
            osatobj = [str(date_repr), str(osat['osat'])]
        return osatobj

    def validate_values(self, hb_osat):
        # Check for sanity on values before saving them
        rc = 0
        errors = []

        if hb_osat:
            if (int(hb_osat) in range(30, 100)):
                hb_osat = int(hb_osat)
            else:
                rc = -1
                errors.append("osat")
        else:
            hb_osat = 0

        if (rc == 0):
            self.set_values(hb_osat)

        else:
            popup = Popup(
                title=tr._('Wrong values'),
                content=Label(
                    text=tr._("Please check {0}").format(errors)),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

    def set_values(self, hb_osat):
        osat = maindb.table('osat')
        current_date = datetime.datetime.now().isoformat()
        domain = 'medical'
        context = 'self_monitoring'

        if hb_osat > 0:
            event_id = str(uuid4())
            synced = False
            osat.insert({'timestamp': current_date,
                         'event_id': event_id,
                         'synced': synced,
                         'osat': hb_osat})

            print("Saved osat", event_id, synced, hb_osat, current_date)

            # Create a new PoL with the values
            # within the medical domain and the self monitoring context
            pol_vals = {
                'page': event_id,
                'page_date': current_date,
                'domain': domain,
                'context': context,
                'measurements': [{'osat': hb_osat}]
            }

            # Create the Page of Life associated to this reading
            PageOfLife.create_pol(PageOfLife, pol_vals)

    @classmethod
    def bol_measurement_formatter_osat(cls, args):
        osat = str(args) or ''
        return tr._("Osat: ") + osat + ' ' + "%"


bol_measurement_formatter.add(
    'osat',
    TrackerBioOsatScreen.bol_measurement_formatter_osat)


class TrackerBioOsatStatsScreen(Screen):
    osat_plot = ObjectProperty()
    date_start = ObjectProperty()
    date_end = ObjectProperty()

    def on_pre_enter(self):
        date_range = default_date_range(
            data=TrackerBioOsatScreen.read_osat()
        )
        self.date_start = date_range[0]
        self.date_end = date_range[1]
        # Update / Refresh the chart anytime we access the stats screen
        self.osat_plot = self.Osatplot()

    def update_pots(self, date_start, date_end):
        self.date_start = date_start
        self.date_end = date_end
        self.osat_plot = self.Osatplot()

    # Plotting - Osat
    def Osatplot(self):
        # Retrieves the history and packages into an array.
        osathist = filter_by_date(
            TrackerBioOsatScreen.read_osat(),
            self.date_start,
            self.date_end)

        osat = []
        osat_date = []
        sorted_list = sorted(osathist, key=lambda sk: sk.get('timestamp'))

        for element in sorted_list:
            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            osat_date.append(dateobj.strftime('%Y-%m-%d'))
            osat.append(element.get('osat'))

        series_osat = {tr._('Oxygen Saturation'): osat}

        chart_io = line_plot(
            title=tr._('Oxygen Saturation') + ' (' + '%' + ')',
            series=series_osat,
            y_legend='%',
            x_values=osat_date,
            renderfmt='png')

        return CoreImage(chart_io, ext="png").texture



##############################################################################
#
#    MyGNUHealth : Mobile and Desktop PHR node for GNU Health
#
#           MyGNUHealth is part of the GNU Health project
#
##############################################################################
#
#    GNU Health: The Libre Digital Health Ecosystem
#    Copyright (C) 2008-2024 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2024 GNU Solidario <health@gnusolidario.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import asyncio
import logging
from bleak import BleakScanner
from mygnuhealth.lang import tr
from mygnuhealth.core import CodeNameMap, maindb
from tinydb import Query
from bleak import BleakClient
from bleak.exc import BleakError

HEART_RATE_UUID = "00002a37-0000-1000-8000-00805f9b34fb"


class BluetoothSettings():

    device_table = maindb.table('devices')

    device_map = CodeNameMap(
        mapinfo=[
            {'code': 'InfiniTime',
             'name': tr.N_('PineTime Smartwatch')},
            {'code': 'btdevice', 'name': tr.N_('Generic')},
        ])

    def get_device_addr(model):
        """Searches and retrieves the device address based on the model"""
        dev = asyncio.run(BluetoothSettings.getdev(model))
        BluetoothSettings.device_address = dev and dev.address or ''
        if dev:
            return dev.address

    async def getdev(model):
        try:
            dev = await BleakScanner.find_device_by_name(model)
        except BleakError:
            logging.info("No bt devices found")
            dev = None
        return dev

    def get_device_info(model):
        """Retrieves the device features from the local database"""
        dev_table = BluetoothSettings.device_table
        res = dev_table.search(Query()['model'] == model)
        if res:
            return res[0]

    def update_device_info(addr, model, features):
        """Updates the ID of the device based on the model"""
        dev_table = BluetoothSettings.device_table

        if (addr and model):
            logging.info(f"Address:{addr}, Model:{model},"
                         f"Features: {features}")

            dev_table.upsert(
                {'model': model,
                 'address': addr,
                 'features': features}, Query().address == addr)

    def get_measures(address, model, feature):
        """Searches and retrieves the device address based on the model"""
        measure = asyncio.run(
            BluetoothSettings.dev_measures(address, model, feature))
        return measure

    async def dev_measures(address, model, feature):
        if feature == 'hr':
            async with BleakClient(address) as client:
                logging.info(f"START: Syncing {feature} for {model}")
                heart_rate = await client.read_gatt_char(HEART_RATE_UUID)
                hr = int.from_bytes(heart_rate)
                logging.info(f"DONE: Value for {feature}: {hr}")
                return hr
        else:
            return None



from kivy.uix.screenmanager import Screen
import datetime
from uuid import uuid4
from kivy.properties import ObjectProperty
from kivy.core.image import Image as CoreImage
from mygnuhealth.core import (maindb, PageOfLife, line_plot,
                              filter_by_date, default_date_range,
                              bol_measurement_formatter)
from mygnuhealth.lang import tr


class TrackerLifestyleSocialScreen(Screen):
    """ Class that manages the person meaningful Social Activity related
        readings. We set and retrieve the latest values from the social
        activity subsystem (meaningful social activities in minutes)

        Attributes:
        -----------

        Methods:
        --------
            set_values: Places the new reading values on the social
                        activity and creates the associated page of life

    """

    # MEANINGFUL SOCIAL ACTIVITIES
    def read_sa():
        # Retrieve the social activity history
        social_activity = maindb.table('socialactivity')
        sahist = social_activity.all()
        return sahist

    def getSA():
        # Extracts the latest readings from Social Activity table
        sahist = TrackerLifestyleSocialScreen.read_sa()
        saobj = ['', '']  # Init to empty string to avoid undefined val
        if sahist:
            sa = sahist[-1]  # Get the latest (newest) record

            dateobj = datetime.datetime.fromisoformat(sa['timestamp'])
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y - %H:%M"))

            saobj = [str(date_repr), str(sa['meaningful'])]

        return saobj

    def set_values(self, meaningful, information):
        social_activity = maindb.table('socialactivity')
        current_date = datetime.datetime.now().isoformat()
        domain = 'lifestyle'
        context = 'social_activities'

        sa_event_id = str(uuid4())
        synced = False
        social_activity.insert(
            {'timestamp': current_date,
             'event_id': sa_event_id,
             'synced': synced,
             'meaningful': meaningful})

        print("Saved Social Activity", sa_event_id, synced,
              meaningful, information)

        # Page of Life block related to Social Activity
        event_id = str(uuid4())
        monitor_readings = [{'sa': {'meaningful': meaningful}}]

        pol_vals = {
            'page': event_id,
            'page_date': current_date,
            'domain': domain,
            'context': context,
            'measurements': monitor_readings,
            'info': information
        }

        # Create the Page of Life associated to this reading
        PageOfLife.create_pol(PageOfLife, pol_vals)

    @classmethod
    def bol_measurement_formatter_sa(cls, args):
        meaningful = str(args.get('meaningful')) or ''
        return (tr._("Meaningful: ") + meaningful)


bol_measurement_formatter.add(
    'sa',
    TrackerLifestyleSocialScreen.bol_measurement_formatter_sa)


class TrackerLifestyleSocialStatsScreen(Screen):
    date_start = ObjectProperty()
    date_end = ObjectProperty()
    sa_plot = ObjectProperty()

    def on_pre_enter(self):
        date_range = default_date_range(
            data=TrackerLifestyleSocialScreen.read_sa()
        )
        self.date_start = date_range[0]
        self.date_end = date_range[1]
        # Update / Refresh the chart anytime we access the stats screen
        self.sa_plot = self.SAplot()

    def update_pots(self, date_start, date_end):
        self.date_start = date_start
        self.date_end = date_end
        self.sa_plot = self.SAplot()

    # Plotting - Social Activity
    def SAplot(self):
        # Retrieves all the history and packages into an array.
        sahist = filter_by_date(
            TrackerLifestyleSocialScreen.read_sa(),
            self.date_start,
            self.date_end)

        sa_meaningful = []
        sa_date = []

        # Sort the list of dictionaries using the timestamp as key
        sorted_list = sorted(sahist, key=lambda sk: sk.get('timestamp'))

        for element in sorted_list:
            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            sa_date.append(dateobj.strftime('%Y-%m-%d'))
            sa_meaningful.append(element.get('meaningful'))

        series_social_activity = {
            tr._('Meaningful Social Activities'): sa_meaningful}

        chart_io = line_plot(
            title=tr._('Meaningful Social Activities') +
            ' (' + tr._('Minutes') + ')',
            series=series_social_activity,
            y_legend=tr._('Minutes'),
            x_values=sa_date,
            renderfmt='png')

        return CoreImage(chart_io, ext="png").texture



####################################################################
#   Copyright (C) 2020-2024 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2024 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import bcrypt
import logging
import datetime
from kivy.uix.label import Label
from kivy.uix.popup import Popup

from mygnuhealth.lang import tr
from mygnuhealth.core import (maindb, get_personal_key,
                              check_date, get_user_profile)


class UserAccount():
    """This class manages the login and initialization of the personal key

        Properties:
        -----------
            login_success: True when entering the correct personal key
            wrong_date: True when an invalid date is found

            account_exist: True when the user account has been created
            today_date: current date

        Methods:
        --------
            init_personal_key: Sets the personal key at the initial run.
            login: Receives the personal key to login, and checks
                if it is the correct one to log in.
            create_account: Receives the initial personal key to
                create an account.

    """

    def get_date(self):
        """
        Returns the date packed into an array (day,month,year)
        """
        rightnow = datetime.datetime.now()
        dateobj = []
        dateobj.append(rightnow.day)
        dateobj.append(rightnow.month)
        dateobj.append(rightnow.year)
        return dateobj

    def account_exist(self):
        """
        Check if an account exist in the database.
        It basically checks for the "credentials" table, created
        at user initialization.
        """
        if (maindb.table('credentials')):
            print("DB is initialized")
            rc = True

        else:
            print("We need to init the personal Key")
            rc = False

        return rc

    def update_profile(self, profile):
        # Include height and other user settings at initialization
        profiletable = maindb.table('profile')
        if (len(profiletable) > 0):
            print(f"Updating profile ... {profile}")
            profiletable.update(profile)

        else:
            print(f"Initializing profile. Setting profile {profile}")
            profiletable.insert(profile)

    def init_personal_key(self, key):
        encrypted_key = bcrypt.hashpw(key.encode('utf-8'),
                                      bcrypt.gensalt()).decode('utf-8')

        credentials_table = maindb.table('credentials')
        if (len(credentials_table) > 0):
            credentials_table.update({'personal_key': encrypted_key})
        else:
            logging.info("Initializing credentials table")
            credentials_table.insert({'personal_key': encrypted_key})

        logging.info(f"Initialized personal key: {encrypted_key}")
        return encrypted_key

    def get_username(self):
        if get_user_profile(maindb):
            return get_user_profile(maindb)['username']

    def login(self, key):
        key = key.encode()

        personal_key = get_personal_key(maindb)

        if bcrypt.checkpw(key, personal_key):
            logging.info("Login correct - Move to main PHR page")
            return True
        else:
            logging.info("Wrong login credentials")
            return False

    def createAccount(self, key, key_repeat,
                      height, personname, birthdate, sex):
        # Retrieves the information from the initialization form
        # Initializes user profile and sets the initial password
        validation_process = True
        errors = []
        profile = {}

        if (height):
            # Sets the user height
            height = int(height)
            profile['height'] = height

        if (personname):
            # Sets the user name
            profile['username'] = personname
        else:
            errors.append("Name")
            validation_process = False

        if (birthdate):
            if (check_date(birthdate)):
                # Sets the birthdate
                year, month, day = birthdate
                daterp = str(datetime.date(int(year), int(month), int(day)))
                profile['birthdate'] = daterp
            else:
                print("Wrong Date!")
                errors.append("Date")
                validation_process = False
                self.wrong_date = True

        # The label -and default value- is "Sex",
        # so we update only if another value is chosen
        if (sex != "Sex"):
            # Sets the user sex
            profile['sex'] = sex
        else:
            errors.append("Sex")
            validation_process = False

        if (key and (key == key_repeat)):
            key = key.encode()
        else:
            errors.append("Personal Key")
            validation_process = False

        if (validation_process and
                self.init_personal_key(key.decode('utf-8'))):
            self.update_profile(profile)

            self.login_success = True
            # If all the info is OK and move to the login screen
            return True

        else:
            popup = Popup(
                title=tr._('Wrong values'),
                content=Label(
                    text=tr._("Please check {0}").format(errors)),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()



from kivy.uix.screenmanager import Screen
import datetime
from uuid import uuid4
from kivy.properties import ObjectProperty
from kivy.core.image import Image as CoreImage
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from mygnuhealth.core import (maindb, PageOfLife, line_plot,
                              filter_by_date, default_date_range,
                              bol_measurement_formatter)
from mygnuhealth.lang import tr


class TrackerBioCardioScreen(Screen):
    """Class that manages the person cardio related readings
        We set and retrieve the latest values from the cardio
        subsystem (blood pressure  and heart rate)

        Attributes:
        -----------

        Methods:
        --------
            set_values: Places the new reading values on the blood
                        pressure and heart rate, and creates the
                        associated page of life

            read_bp: Retrieves the BP history
            read_hr: Retrieves the HR history
            getBP: Extracts the latest readings from BP
    """

    def read_bp():
        # Retrieves the BP history
        blood_pressure = maindb.table('bloodpressure')
        bphist = blood_pressure.all()
        return bphist

    def read_hr():
        # Retrieves the HR history
        hr = maindb.table('heart_rate')
        hrhist = hr.all()
        return (hrhist)

    def getBP():
        # Extracts the latest readings from BP
        bphist = TrackerBioCardioScreen.read_bp()
        hrhist = TrackerBioCardioScreen.read_hr()
        bpobj = ['', '', '', '']  # Init to empty string to avoid undefined val
        if bphist and hrhist:
            bp = bphist[-1]  # Get the latest (newest) record
            hr = hrhist[-1]
            dateobj = datetime.datetime.fromisoformat(bp['timestamp'])
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y - %H:%M"))

            bpobj = [str(date_repr), str(bp['systolic']), str(bp['diastolic']),
                     str(hr['heart_rate'])]

        return bpobj

    def validate_values(self, systolic, diastolic, heart_rate):
        # Check for sanity on values before saving them
        rc = 0
        errors = []
        if systolic:
            if (int(systolic) in range(20, 300)):
                systolic = int(systolic)
            else:
                rc = -1
                errors.append("Systolic")
        else:
            systolic = 0

        if diastolic:
            if (int(diastolic) in range(20, 300)):
                diastolic = int(diastolic)
            else:
                errors.append("Diastolic")
                rc = -1
        else:
            diastolic = 0

        if heart_rate:
            if (int(heart_rate) in range(20, 400)):
                heart_rate = int(heart_rate)
            else:
                print("Wrong value for heart_rate")
                rc = -1
                errors.append("Heart rate")
        else:
            heart_rate = 0
            print("No heart rate")

        if (rc == 0):
            self.set_values(systolic, diastolic, heart_rate)

        else:
            popup = Popup(
                title=tr._('Wrong values'),
                content=Label(
                    text=tr._("Please check {0}").format(errors)),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

    def set_values(self, systolic, diastolic, heart_rate, extra_info=None):
        blood_pressure = maindb.table('bloodpressure')
        hr = maindb.table('heart_rate')
        current_date = datetime.datetime.now().isoformat()
        bpmon = hrmon = False  # Init to false the bp and hr monitoring process
        domain = 'medical'
        context = 'self_monitoring'

        if (systolic > 0) and (diastolic > 0):
            bpmon = True
            bp_event_id = str(uuid4())
            synced = False
            blood_pressure.insert({'timestamp': current_date,
                                   'event_id': bp_event_id,
                                   'synced': synced,
                                   'systolic': systolic,
                                   'diastolic': diastolic})

            print("Saved blood pressure", bp_event_id, synced, systolic,
                  diastolic, current_date)

        if heart_rate > 0:
            hrmon = True
            hr_event_id = str(uuid4())
            synced = False
            hr.insert({'timestamp': current_date,
                       'event_id': hr_event_id,
                       'synced': synced,
                       'heart_rate': heart_rate})

            print("Saved Heart rate", hr_event_id, synced,
                  heart_rate, current_date)

        if (bpmon or hrmon):
            # This block is related to the Page of Life creation
            if (bpmon and hrmon):
                # Group both HR and BP monitors in one PoL if both readings
                # where taken at the same moment / device
                # The event_id will be unique
                event_id = str(uuid4())
                # Used to generate po items.
                monitor_readings = [
                    {'bp': {'systolic': systolic, 'diastolic': diastolic}},
                    {'hr': heart_rate}
                ]
            elif (bpmon and not hrmon):
                event_id = bp_event_id
                monitor_readings = [
                    {'bp': {'systolic': systolic, 'diastolic': diastolic}},
                ]
            else:
                event_id = hr_event_id
                monitor_readings = [{'hr': heart_rate}]

            pol_vals = {
                'page': event_id,
                'page_date': current_date,
                'domain': domain,
                'context': context,
                'measurements': monitor_readings,
                'info': extra_info
            }

            # Create the Page of Life associated to this reading
            PageOfLife.create_pol(PageOfLife, pol_vals)

    @classmethod
    def bol_measurement_formatter_bp(cls, args):
        systolic = str(args.get('systolic')) or ''
        diastolic = str(args.get('diastolic')) or ''

        return tr._("BP: ") + f"{systolic} / {diastolic} " + tr._("mmHg")

    @classmethod
    def bol_measurement_formatter_hr(cls, args):
        hr = str(args) or ''

        return tr._("Heart Rate: ") + hr + ' ' + tr._("bpm")


bol_measurement_formatter.add(
    'bp', TrackerBioCardioScreen.bol_measurement_formatter_bp)


bol_measurement_formatter.add(
    'hr', TrackerBioCardioScreen.bol_measurement_formatter_hr)


class TrackerBioCardioStatsScreen(Screen):
    hr_plot = ObjectProperty()
    bp_plot = ObjectProperty()
    date_start = ObjectProperty()
    date_end = ObjectProperty()

    def on_pre_enter(self):
        date_range = default_date_range(
            data=TrackerBioCardioScreen.read_hr()
        )
        self.date_start = date_range[0]
        self.date_end = date_range[1]

        # Update / Refresh the chart anytime we access the stats screen
        self.hr_plot = self.HRplot()
        self.bp_plot = self.BPplot()

    def update_pots(self, date_start, date_end):
        self.date_start = date_start
        self.date_end = date_end
        self.hr_plot = self.HRplot()
        self.bp_plot = self.BPplot()

    # Plotting - Heart rate
    def HRplot(self):
        # Retrieves all the history and packages into an array.
        hrhist = filter_by_date(
            TrackerBioCardioScreen.read_hr(),
            self.date_start,
            self.date_end)
        hr = []
        hr_date = []
        sorted_list = sorted(hrhist, key=lambda sk: sk.get('timestamp'))

        for element in sorted_list:
            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            hr_date.append(dateobj.strftime('%Y-%m-%d'))
            hr.append(element.get('heart_rate'))

        series_hr = {tr._('Heart Rate'): hr}

        # chart_io = line_plot('Heart rate', 'BPM', hr, x_values=None)
        chart_io = line_plot(
            title=tr._('Heart rate') + ' (' + tr._('BPM') + ')',
            series=series_hr,
            y_legend=tr._('BPM'),
            x_values=hr_date,
            renderfmt='png')

        return CoreImage(chart_io, ext="png").texture

    # Plotting - Blood Pressure
    def BPplot(self):
        # Retrieves all the history and packages into an array.
        bphist = filter_by_date(
            TrackerBioCardioScreen.read_bp(),
            self.date_start,
            self.date_end)
        bpsys = []
        bpdia = []
        bp_date = []

        # Sort the list of dictionaries using the timestamp as key
        sorted_list = sorted(bphist, key=lambda sk: sk.get('timestamp'))

        for element in sorted_list:

            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            bp_date.append(dateobj.strftime('%Y-%m-%d'))
            bpsys.append(element.get('systolic'))
            bpdia.append(element.get('diastolic'))

        series_bp = {tr._('Systolic'): bpsys,
                     tr._('Diastolic'): bpdia}

        chart_io = line_plot(
            title=tr._('Blood Pressure') + ' (' + tr._('mmHg') + ')',
            series=series_bp,
            y_legend=tr._('mmHg'),
            x_values=bp_date,
            renderfmt='png')

        return CoreImage(chart_io, ext="png").texture



from kivy.uix.screenmanager import Screen
import datetime
from uuid import uuid4
from kivy.properties import ObjectProperty
from kivy.core.image import Image as CoreImage
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from mygnuhealth.core import (maindb, PageOfLife, line_plot, CodeNameMap,
                              filter_by_date, default_date_range,
                              bol_measurement_formatter)
from mygnuhealth.lang import tr


class TrackerLifestyleSleepScreen(Screen):
    """ Class that manages the person Sleep related readings
        We set and retrieve the latest values from the sleep
        subsystem (time and quality)

        Attributes:
        -----------

        Methods:
        --------
            set_values: Places the new reading values on the sleep
                        and creates the associated page of life

    """
    # SLEEP
    def read_sleep():
        # Retrieve the sleep history
        sleep = maindb.table('sleep')
        sleephist = sleep.all()
        return sleephist

    @classmethod
    def getSleep(cls):
        # Extracts the latest readings from the sleep table
        sleephist = TrackerLifestyleSleepScreen.read_sleep()
        # date, hours, sleep quality
        sleepobj = ['', '', '']
        if sleephist:
            sleep = sleephist[-1]  # Get the latest (newest) record

            dateobj = datetime.datetime.fromisoformat(sleep['timestamp'])
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y - %H:%M"))

            sleepobj = [str(date_repr), str(sleep['sleeptime']),
                        str(cls.get_sleep_quality_name(
                            sleep['sleepquality']))]

        return sleepobj

    sleep_quality = CodeNameMap(
        mapinfo=[
            {'code': 'good', 'name': tr.N_('Good')},
            {'code': 'light', 'name': tr.N_('Light')},
            {'code': 'poor', 'name': tr.N_('Poor')},
        ])

    @classmethod
    def get_sleep_quality_name(cls, sleep_quality_code):
        return cls.sleep_quality.get_name(sleep_quality_code)

    def validate_values(self, sleeptime, sleepquality_name, information):
        # Check for sanity on values before saving them
        rc = 0
        errors = []
        sleepquality_code = self.sleep_quality.get_code(sleepquality_name)

        if sleeptime:
            if (0.1 <= float(sleeptime) <= 23.5):
                sleeptime = float(sleeptime)
            else:
                rc = -1
                errors.append("Sleep time")
        else:
            sleeptime = 0

        if (rc == 0):
            self.set_values(sleeptime, sleepquality_code, information)

        else:
            popup = Popup(
                title=tr._('Wrong values'),
                content=Label(
                    text=tr._("Please check {0}").format(errors)),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

    def set_values(self, sleeptime, sleepquality_code, information):
        sleep = maindb.table('sleep')
        current_date = datetime.datetime.now().isoformat()
        domain = 'lifestyle'
        context = 'sleep'

        sleep_event_id = str(uuid4())
        synced = False
        sleep.insert({'timestamp': current_date,
                      'event_id': sleep_event_id,
                      'synced': synced,
                      'sleeptime': sleeptime,
                      'sleepquality': sleepquality_code})

        print("Saved Sleep information", sleep_event_id,
              sleeptime, sleepquality_code, current_date)

        # Page of Life block related to Sleep
        event_id = str(uuid4())
        monitor_readings = [
            {'sleep': {'sleeptime': sleeptime,
                       'sleepquality': sleepquality_code}},
        ]

        pol_vals = {
            'page': event_id,
            'page_date': current_date,
            'domain': domain,
            'context': context,
            'measurements': monitor_readings,
            'info': information
        }

        # Create the Page of Life associated to this reading
        PageOfLife.create_pol(PageOfLife, pol_vals)

    @classmethod
    def bol_measurement_formatter_sleep(cls, args):
        sleeptime = str(args.get('sleeptime')) or ''
        sleepquality_code = str(args.get('sleepquality')) or ''
        sleepquality_name = cls. get_sleep_quality_name(
            sleepquality_code) or ''

        return (tr._("Sleep Time: ") + sleeptime + '\n' +
                tr._("Sleep Quality: ") + sleepquality_name)


bol_measurement_formatter.add(
    'sleep',
    TrackerLifestyleSleepScreen.bol_measurement_formatter_sleep)


class TrackerLifestyleSleepStatsScreen(Screen):
    date_start = ObjectProperty()
    date_end = ObjectProperty()
    sleep_plot = ObjectProperty()

    def on_pre_enter(self):
        date_range = default_date_range(
            data=TrackerLifestyleSleepScreen.read_sleep()
        )
        self.date_start = date_range[0]
        self.date_end = date_range[1]
        # Update / Refresh the chart anytime we access the stats screen
        self.sleep_plot = self.Sleepplot()

    def update_pots(self, date_start, date_end):
        self.date_start = date_start
        self.date_end = date_end
        self.sleep_plot = self.Sleepplot()

    # Plotting - Sleep
    def Sleepplot(self):
        # Retrieves all the history and packages into an array.
        sleephist = filter_by_date(
            TrackerLifestyleSleepScreen.read_sleep(),
            self.date_start,
            self.date_end)

        sleep_time = []
        sleep_date = []

        # Sort the list of dictionaries using the timestamp as key
        sorted_list = sorted(sleephist, key=lambda sk: sk.get('timestamp'))

        for element in sorted_list:
            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            sleep_date.append(dateobj.strftime('%Y-%m-%d'))
            sleep_time.append(element.get('sleeptime'))

        series_sleep = {tr._('Sleep'): sleep_time}

        chart_io = line_plot(
            title=tr._('Sleep') + ' (' + tr._('Hours') + ')',
            series=series_sleep,
            y_legend=tr._('Hours'),
            x_values=sleep_date,
            renderfmt='png')

        return CoreImage(chart_io, ext="png").texture



from kivy.uix.screenmanager import Screen
import datetime
from uuid import uuid4
from kivy.properties import ObjectProperty
from kivy.core.image import Image as CoreImage
from mygnuhealth.core import (maindb, PageOfLife, line_plot,
                              filter_by_date, default_date_range,
                              bol_measurement_formatter)
from mygnuhealth.lang import tr


class TrackerPsychoMoodScreen(Screen):
    """Class that manages the person psychological related readings
        We set and retrieve the latest values from the psycho
        subsystem (mood and energy)

        Attributes:
        -----------

        Methods:
        --------
            set_values: Places the new reading values on the mood
                        and energy, and creates the
                        associated page of life

            read_mood: Retrieves the BP history
            getMood: Extracts the latest readings from the mood table
    """

    def read_mood():
        # Retrieve the  history
        mood_table = maindb.table('mood')
        moodhist = mood_table.all()
        return moodhist

    def getMood():
        # Extracts the latest readings from mood table (both mood and energy)
        moodhist = TrackerPsychoMoodScreen.read_mood()
        moodobj = ['', '', '']  # Init to empty string to avoid undefined val
        if moodhist:
            mood = moodhist[-1]  # Get the latest (newest) record
            dateobj = datetime.datetime.fromisoformat(mood['timestamp'])
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y - %H:%M"))
            moodobj = [str(date_repr), str(mood['mood']), str(mood['energy'])]

        return moodobj

    def set_values(self, mood, energy, information):
        mood_table = maindb.table('mood')
        current_date = datetime.datetime.now().isoformat()
        moodmon = False  # Init to false the mood monitoring process
        domain = 'medical'
        context = 'self_monitoring'

        if (energy > -1):  # Will evaluate to True since energy lower lim = 0
            moodmon = True
            mood_event_id = str(uuid4())
            synced = False
            mood_table.insert({'timestamp': current_date,
                               'event_id': mood_event_id,
                               'synced': synced,
                               'mood': mood,
                               'energy': energy})

            print("Saved Mood and Energy Levels", mood_event_id, synced, mood,
                  energy, current_date)

        if (moodmon):
            # This block is related to the Page of Life creation
            event_id = str(uuid4())
            monitor_readings = [
                {'mood_energy': {'mood': mood,
                                 'energy': energy}},
            ]

            pol_vals = {
                'page': event_id,
                'page_date': current_date,
                'domain': domain,
                'context': context,
                'measurements': monitor_readings,
                'info': information
            }

            # Create the Page of Life associated to this reading
            PageOfLife.create_pol(PageOfLife, pol_vals)

    @classmethod
    def bol_measurement_formatter_mood_energy(cls, args):
        mood = str(args.get('mood')) or ''
        energy = str(args.get('energy')) or ''

        return (tr._("Mood: ") + mood + '\n' +
                tr._("Energy: ") + energy)


bol_measurement_formatter.add(
    'mood_energy',
    TrackerPsychoMoodScreen.bol_measurement_formatter_mood_energy)


class TrackerPsychoMoodStatsScreen(Screen):
    date_start = ObjectProperty()
    date_end = ObjectProperty()
    mood_plot = ObjectProperty()
    energy_plot = ObjectProperty()

    def on_pre_enter(self):
        date_range = default_date_range(
            data=TrackerPsychoMoodScreen.read_mood()
        )
        self.date_start = date_range[0]
        self.date_end = date_range[1]
        # Update / Refresh the chart anytime we access the stats screen
        self.mood_plot, self.energy_plot = self.Moodplot()

    def update_pots(self, date_start, date_end):
        self.date_start = date_start
        self.date_end = date_end
        self.mood_plot, self.energy_plot = self.Moodplot()

    # Plotting - Mood and Energy
    def Moodplot(self):
        # Retrieves all the history and packages into an array.
        moodhist = filter_by_date(
            TrackerPsychoMoodScreen.read_mood(),
            self.date_start,
            self.date_end)

        mood = []
        energy = []
        mood_date = []

        # Sort the list of dictionaries using the timestamp as key
        sorted_list = sorted(moodhist, key=lambda sk: sk.get('timestamp'))

        for element in sorted_list:
            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            mood_date.append(dateobj.strftime('%Y-%m-%d'))
            mood.append(element.get('mood'))
            energy.append(element.get('energy'))

        series_mood = {tr._('Mood'): mood}
        series_energy = {tr._('Energy'): energy}

        chart_mood = line_plot(
            title=tr._("Mood"),
            series=series_mood,
            y_legend=tr._('Mood'),
            x_values=mood_date,
            renderfmt='png')

        chart_energy = line_plot(
            title=tr._("Energy"),
            series=series_energy,
            y_legend=tr._('Energy'),
            x_values=mood_date,
            renderfmt='png')

        return [CoreImage(chart_mood, ext="png").texture,
                CoreImage(chart_energy, ext="png").texture]



import json
import requests
import threading
import datetime
from tinydb import Query

from mygnuhealth.core import (maindb, boldb,
                              bol_domain_formatter,
                              bol_measurement_formatter,
                              bol_genetic_info_formatter)

from mygnuhealth.lang import tr


class BookofLife():
    """Class that manages the person Book of Life

        Attributes:
        -----------
            boldb: TinyDB instance.
                Holds the book of life with all the events (pages of life)
        Methods:
        --------
            read_book: retrieves all pages
            format_bol: compacts and shows the relevant fields in a
            human readable format
    """

    def format_bol(bookoflife):
        """Takes the necessary fields and formats the book in a way that can
        be shown in the device, mixing fields and compacting entries in a more
        human readable format"""
        book = []
        for pageoflife in bookoflife:
            pol = {}
            summ = []
            msr = []

            # Page ID
            pol['page_id'] = pageoflife['page']

            # Federation sync state.
            pol['fsynced'] = pageoflife['fsynced']

            # Privacy page
            pol['privacy'] = pageoflife['privacy']

            # Date
            dateobj = datetime.datetime.fromisoformat(pageoflife['page_date'])
            # Use a localized and easy to read date format
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y\n%H:%M"))
            pol['date'] = date_repr

            # Domain and context
            pol['domain'] = bol_domain_formatter.handle(
                'domain', [pageoflife['domain'], pageoflife['context']])

            # Title
            title = pageoflife['summary']
            if title:
                summ.append(f'{title}')

            # Measurement
            mvals = pageoflife['measurements']
            if ('measurements' in pageoflife.keys() and
                    mvals):

                for measure in mvals:
                    for name, values in measure.items():
                        msr.append(
                            bol_measurement_formatter.handle(
                                name, values))

                summ = summ + msr

            # Genetics
            genetics = pageoflife['genetic_info']
            if ('genetic_info' in pageoflife.keys() and
                    genetics):

                summ.append(
                    bol_genetic_info_formatter.handle(
                        'genetic_info', genetics))

            # Detail info
            details = pageoflife['info']
            if details:
                summ.append(f'{details}')

            pol['summary'] = "\n".join(summ)
            book.append(pol)
        return book

    def read_book():
        """retrieves all pages of the individual Book of Life
        """
        booktable = boldb.table('pol')
        book = booktable.all()
        formatted_bol = BookofLife.format_bol(book)
        return formatted_bol

    def check_sync_status(self):
        fedinfo = maindb.table('federation')
        if len(fedinfo):
            sync = fedinfo.all()[0]['enable_sync']
            return sync

    def sync_book(fedkey, callback=None):
        # Emit the signal to display a busy indicator while
        # pushing the pages of life to the GH federation
        print(tr._("***** Initiating the synchronization "
                   "with the GH federation server"))
        # self.pushingPols.emit()
        # Spawn a new thread so synchronization / pushing is done
        # asynchronously in a non-blocking fashion
        thread = threading.Thread(
            name="pushpols_thread",
            target=BookofLife.push_pols,
            args=(fedkey, callback))
        thread.start()

    def push_pols(fedkey, callback=None):
        state = {}

        try:
            BookofLife._push_pols(fedkey)
        except BaseException as e:
            print(f'Sync error: {e}')
            state['error'] = e

        if callback:
            callback(state)

    def _push_pols(fedkey):
        """This method will go through each page in the book of life
        that has not been sent to the GNU Health Federation server yet
        (fsynced = False).
        It also checks for records that have a book associated to it
        and that the specific page is has not the "private" flag set.

        Parameters
        ----------
        """
        fedinfo = maindb.table('federation')
        if not len(fedinfo):
            raise Exception(
                tr._("Thalamus settings have problems in network setting."))
            return

        res = fedinfo.all()[0]

        # Refresh all pages of life
        booktable = boldb.table('pol')
        book = booktable.all()
        user = res.get('federation_account')
        protocol = res.get('protocol')
        server = res.get('federation_server')
        port = res.get('federation_port')

        if not user or user == '':
            raise Exception(
                tr._("Can't find Federation account from profile setting."))
            return

        if not fedkey or fedkey == '':
            raise Exception(
                tr._("Need to enter the federation password."))
            return

        pages_send_failed = []

        for pol in book:
            timestamp = pol.get('page_date')
            node = pol.get('node')
            page_id = pol.get('page')
            synced = pol.get('fsynced')

            # Only sync those pages that are not private
            privacy = pol.get('privacy')
            if not privacy and not synced:
                creation_info = {
                    'user': user,
                    'timestamp': timestamp,
                    'node': node}

                pol['creation_info'] = creation_info
                pol['id'] = page_id

                url = f"{protocol}://{server}:{port}/pols/{user}/{page_id}"

                pol['fsynced'] = True
                send_data = requests.request(
                    'POST', url,
                    data=json.dumps(pol),
                    auth=(user, fedkey),
                    verify=False)

                if send_data:
                    Page = Query()
                    booktable.update(
                        {'fsynced': True}, Page.page == page_id)
                else:
                    pages_send_failed.append(page_id)

        if len(pages_send_failed) > 0:
            raise Exception(
                tr._("Pages send failed: {}").format(
                    '\n'.join(pages_send_failed[0:1000])))



from kivy.uix.screenmanager import Screen
from kivy.properties import ObjectProperty
from kivy.core.image import Image as CoreImage
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from mygnuhealth.core import (maindb, PageOfLife, line_plot,
                              filter_by_date, default_date_range,
                              bol_measurement_formatter)
from mygnuhealth.lang import tr

import datetime
from uuid import uuid4

# In a future, try to work directly with SVG format generated from pygal
# from kivy.graphics.svg import Svg
# from kivy.uix.scatter import Scatter

"""
# For future use when pygal svg chart better integrates with Kivy
class SvgWidget(Scatter):

    def __init__(self, **kwargs):
        super(SvgWidget, self).__init__(**kwargs)
        with self.canvas:
            src = TrackerBioWeightStatsScreen.Weightplot()
            svg = Svg(source=src)
"""


class TrackerBioWeightScreen(Screen):
    """Class that manages the person weight readings

        Attributes:
        -----------

        Methods:
        --------
            set_values: Places the new reading values on the 'weight'
            and creates the associated page of life
            default_weight: Gets the latest weight (TODO replace by getWeight)
            read_weight: Retrieve the weight levels history
            getWeight: Extracts the latest readings from Weight
    """

    def default_weight(self):
        weighttable = maindb.table('weight')
        if (len(weighttable) > 0):
            last_weight = weighttable.all()[-1]
            return (last_weight['weight'])
        else:
            return 0

    def read_weight():
        # Retrieve the weight levels history
        weighttable = maindb.table('weight')
        weighthist = weighttable.all()
        return (weighthist)

    def getWeight():
        # Extracts the latest readings from Weight
        weighthist = TrackerBioWeightScreen.read_weight()
        weightobj = ['', '']
        if (weighthist):
            weight = weighthist[-1]  # Get the latest (newest) record
            dateobj = datetime.datetime.fromisoformat(weight['timestamp'])
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y - %H:%M"))

            weightobj = [str(date_repr), str(weight['weight'])]
        return weightobj

    def validate_values(self, body_weight):
        # Check for sanity on values before saving them
        rc = 0
        errors = []

        if body_weight:
            if (2 <= float(body_weight) < 500):
                body_weight = float(body_weight)
            else:
                print("Wrong value for weight")
                rc = -1
                errors.append("Body weight")
        else:
            body_weight = 0
            print("No weight")

        if (rc == 0):
            self.set_values(body_weight)

        else:
            popup = Popup(
                title=tr._('Wrong values'),
                content=Label(
                    text=tr._("Please check {0}").format(errors)),
                size_hint=(0.5, 0.5), auto_dismiss=True)

            popup.open()

    def set_values(self, body_weight):
        weighttable = maindb.table('weight')
        profiletable = maindb.table('profile')
        current_date = datetime.datetime.now().isoformat()
        domain = 'medical'
        context = 'self_monitoring'

        if body_weight > 0:
            event_id = str(uuid4())
            synced = False
            height = None
            bmi = None
            if (len(profiletable) > 0):
                height = profiletable.all()[0]['height']
            vals = {'timestamp': current_date,
                    'event_id': event_id,
                    'synced': synced,
                    'weight': body_weight}
            measurements = {'wt': body_weight}

            # If height is in the person profile, calculate the BMI
            if height:
                bmi = body_weight / ((height / 100)**2)
                bmi = round(bmi, 1)  # Use one decimal
                vals['bmi'] = bmi
                measurements['bmi'] = bmi

            weighttable.insert(vals)

            print("Saved weight", event_id, synced, body_weight, bmi,
                  current_date)

            # Create a new PoL with the values
            # within the medical domain and the self monitoring context
            pol_vals = {
                'page': event_id,
                'page_date': current_date,
                'domain': domain,
                'context': context,
                'measurements': [measurements]
            }

            # Create the Page of Life associated to this reading
            PageOfLife.create_pol(PageOfLife, pol_vals)

    @classmethod
    def bol_measurement_formatter_weight(cls, args):
        weight = str(args) or ''
        return tr._("Weight: ") + weight + ' ' + tr._("kg")

    @classmethod
    def bol_measurement_formatter_bmi(cls, args):
        bmi = str(args) or ''
        return tr._("BMI: ") + bmi + ' ' + tr._("kg/m2")


bol_measurement_formatter.add(
    'wt',
    TrackerBioWeightScreen.bol_measurement_formatter_weight)


bol_measurement_formatter.add(
    'bmi',
    TrackerBioWeightScreen.bol_measurement_formatter_bmi)


class TrackerBioWeightStatsScreen(Screen):
    date_start = ObjectProperty()
    date_end = ObjectProperty()
    weight_plot = ObjectProperty(None)

    def on_pre_enter(self):
        date_range = default_date_range(
            data=TrackerBioWeightScreen.read_weight()
        )
        self.date_start = date_range[0]
        self.date_end = date_range[1]
        # Update / Refresh the chart anytime we access the stats screen
        self.weight_plot = self.Weightplot()

    def update_pots(self, date_start, date_end):
        self.date_start = date_start
        self.date_end = date_end
        self.weight_plot = self.Weightplot()

    # Plotting - Weight
    def Weightplot(self):
        # Retrieves the history and packages into an array.
        weighthist = filter_by_date(
            TrackerBioWeightScreen.read_weight(),
            self.date_start,
            self.date_end)

        weight = []
        weight_date = []
        sorted_list = sorted(weighthist, key=lambda sk: sk['timestamp'])

        for element in sorted_list:
            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            # Transform datetime obj to str for the x labels
            weight_date.append(dateobj.strftime('%Y-%m-%d'))
            weight.append(element.get('weight'))

        series_weight = {tr._('Weight'): weight}

        # In the future we could use renderfmt='svg' instead of png
        chart_io = line_plot(
            title=tr._('Weight') + ' (' + tr._('Kg') + ')',
            series=series_weight,
            y_legend=tr._('Kg'),
            x_values=weight_date,
            renderfmt='png')

        return CoreImage(chart_io, ext='png').texture

        # return chart  # Use for future SVG plots



from kivy.uix.screenmanager import Screen
import datetime
from uuid import uuid4
from kivy.properties import ObjectProperty
from kivy.core.image import Image as CoreImage
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from mygnuhealth.core import (maindb, PageOfLife, stacked_plot,
                              filter_by_date, default_date_range,
                              bol_measurement_formatter)
from mygnuhealth.lang import tr


class TrackerLifestyleNutritionScreen(Screen):
    """ Class that manages the person Nutrition related readings
        We set and retrieve the latest values from the nutrition
        subsystem (calorie intake at different times of the day)

        Attributes:
        -----------

        Methods:
        --------
            set_values: Places the new reading values on the nutrition
                        and creates the associated page of life

    """

    # NUTRITION
    def read_nutrition():
        # Retrieve the nutrition history
        nutrition = maindb.table('nutrition')
        nutrihist = nutrition.all()
        return nutrihist

    def getNutrition():
        # Extracts the latest readings from the nutrition table
        nutrihist = TrackerLifestyleNutritionScreen.read_nutrition()
        # Init to empty string to avoid undefined val
        nutriobj = ['', '', '', '', '']
        if nutrihist:
            nutri = nutrihist[-1]  # Get the latest (newest) record

            dateobj = datetime.datetime.fromisoformat(nutri['timestamp'])
            date_repr = dateobj.strftime(tr._("%a, %b %d '%y - %H:%M"))

            nutriobj = [str(date_repr), str(nutri['calmorning']),
                        str(nutri['calafternoon']),
                        str(nutri['calevening']),
                        str(nutri['caltotal'])
                        ]

        return nutriobj

    def validate_values(self, calmorning, calafternoon,
                        calevening, caltotal, information):
        # Check for sanity on values before saving them
        rc = 0
        errors = []
        if calmorning:
            if (int(calmorning) in range(1, 50000)):
                calmorning = int(calmorning)
            else:
                rc = -1
                errors.append("Morning")
        else:
            calmorning = 0

        if calafternoon:
            if (int(calafternoon) in range(1, 50000)):
                calafternoon = int(calafternoon)
            else:
                errors.append("Afternoon")
                rc = -1
        else:
            calafternoon = 0

        if calevening:
            if (int(calevening) in range(1, 50000)):
                calevening = int(calevening)
            else:
                rc = -1
                errors.append("Evening")
        else:
            calevening = 0

        if caltotal:
            if (int(caltotal) in range(1, 50000)):
                calevening = int(calevening)
            else:
                rc = -1
                errors.append("Total")
        else:
            calevening = 0

        if (rc == 0):
            self.set_values(calmorning, calafternoon, calevening,
                            caltotal, information)

        else:
            popup = Popup(
                title=tr._('Wrong values'),
                content=Label(
                    text=tr._("Please check {0}").format(errors)),
                size_hint=(0.5, 0.5), auto_dismiss=True)
            popup.open()

    def set_values(self, calmorning, calafternoon, calevening, caltotal,
                   information):
        nutrition = maindb.table('nutrition')
        current_date = datetime.datetime.now().isoformat()
        domain = 'lifestyle'
        context = 'nutrition'

        nutrition_event_id = str(uuid4())
        synced = False
        nutrition.insert({'timestamp': current_date,
                          'event_id': nutrition_event_id,
                          'synced': synced,
                          'calmorning': calmorning,
                          'calafternoon': calafternoon,
                          'calevening': calevening,
                          'caltotal': caltotal})

        print("Saved Nutrition information", nutrition_event_id,
              calmorning, calafternoon, calevening, caltotal, current_date)

        # Page of Life block related to Nutrition
        event_id = str(uuid4())

        monitor_readings = [
            {'nutrition': {'calmorning': calmorning,
                           'calafternoon': calafternoon,
                           'calevening': calevening,
                           'caltotal': caltotal}},
        ]

        pol_vals = {
            'page': event_id,
            'page_date': current_date,
            'domain': domain,
            'context': context,
            'measurements': monitor_readings,
            'info': information
        }

        # Create the Page of Life associated to this reading
        PageOfLife.create_pol(PageOfLife, pol_vals)

    @classmethod
    def bol_measurement_formatter_nutrition(cls, args):
        calmorning = str(args.get('calmorning')) or ''
        calafternoon = str(args.get('calafternoon')) or ''
        calevening = str(args.get('calevening')) or ''
        caltotal = str(args.get('caltotal')) or ''
        u = tr._('Kcal')

        return (tr._("Morning: ") + calmorning + ' ' + u + ',  ' +
                tr._("Afternoon: ") + calafternoon + ' ' + u + '\n' +
                tr._("Evening: ") + calevening + ' ' + u + ',  ' +
                tr._("Total: ") + caltotal + ' ' + u)


bol_measurement_formatter.add(
    'nutrition',
    TrackerLifestyleNutritionScreen.bol_measurement_formatter_nutrition)


class TrackerLifestyleNutritionStatsScreen(Screen):
    date_start = ObjectProperty()
    date_end = ObjectProperty()
    nutrition_plot = ObjectProperty()

    def on_pre_enter(self):
        date_range = default_date_range(
            data=TrackerLifestyleNutritionScreen.read_nutrition()
        )
        self.date_start = date_range[0]
        self.date_end = date_range[1]
        # Update / Refresh the chart anytime we access the stats screen
        self.nutrition_plot = self.Nutritionplot()

    def update_pots(self, date_start, date_end):
        self.date_start = date_start
        self.date_end = date_end
        self.nutrition_plot = self.Nutritionplot()

    def convert(self, value):
        try:
            value = float(value)
        except BaseException:
            value = 0

        return value

    # Plotting - Nutrition
    def Nutritionplot(self):
        # Retrieves all the history and packages into an array.
        nutrihist = filter_by_date(
            TrackerLifestyleNutritionScreen.read_nutrition(),
            self.date_start,
            self.date_end)
        nutri_morning = []
        nutri_afternoon = []
        nutri_evening = []
        nutri_date = []

        # Sort the list of dictionaries using the timestamp as key
        sorted_list = sorted(nutrihist, key=lambda sk: sk.get('timestamp'))

        for element in sorted_list:
            dateobj = datetime.datetime.fromisoformat(element.get('timestamp'))
            nutri_date.append(dateobj.strftime('%Y-%m-%d'))
            nutri_morning.append(self.convert(element.get('calmorning')))
            nutri_afternoon.append(self.convert(element.get('calafternoon')))
            nutri_evening.append(self.convert(element.get('calevening')))

        series = {tr._('Morning'): nutri_morning,
                  tr._('Afternoon'): nutri_afternoon,
                  tr._('Evening'): nutri_evening}

        chart_io = stacked_plot(
            title=tr._('Nutrition') + ' (' + tr._('Kcal') + ')',
            series=series,
            x_values=nutri_date,
            renderfmt='png')

        return CoreImage(chart_io, ext="png").texture
